#!/usr/bin/env python
"""
QEMU_SERIAL Communication Interface to communicate with emulated QEMU_SERIAL hardware via the
UART interface.

It utilizes the the asyncio library.

Requirements:
  Python >= 3.7 (asyncio support)

Instructions:
  Run QEMU_SERIAL (modified for OBSW) via

  qemu-system-arm -M isis-obc -monitor stdio \
      -bios path/to/sourceobsw-at91sam9g20_ek-sdram.bin \
      -qmp unix:/tmp/qemu,server -S

  Then run the telecommand script with -c 2
"""

import asyncio
import logging
import struct
import json
import re
import errno
import sys
import time
from collections import deque
from threading import Thread
from typing import Optional

from tmtccmd.com import ComInterface
from tmtccmd.tmtc import TelemetryListT
from tmtccmd.com.serial_base import SerialCfg, SerialCommunicationType
from dle_encoder import DleEncoder, STX_CHAR, ETX_CHAR, DleErrorCodes

_LOGGER = logging.getLogger(__name__)
SERIAL_FRAME_LENGTH = 256
DLE_FRAME_LENGTH = 1500

# Paths to Unix Domain Sockets used by the emulator
QEMU_ADDR_QMP = "/tmp/qemu"

# Request/response category and command IDs
IOX_CAT_DATA = 0x01
IOX_CAT_FAULT = 0x02

IOX_CID_DATA_IN = 0x01
IOX_CID_DATA_OUT = 0x02

IOX_CID_FAULT_OVRE = 0x01
IOX_CID_FAULT_FRAME = 0x02
IOX_CID_FAULT_PARE = 0x03
IOX_CID_FAULT_TIMEOUT = 0x04


def start_background_loop(loop: asyncio.AbstractEventLoop) -> None:
    asyncio.set_event_loop(loop)
    loop.run_forever()


class QEMUComIF(ComInterface):
    """Specific Communication Interface implementation of the QEMU_SERIAL USART protocol for the
    TMTC software"""

    def __init__(
        self,
        serial_cfg: SerialCfg,
        ser_com_type: SerialCommunicationType = SerialCommunicationType.DLE_ENCODING,
    ):
        self.cfg = serial_cfg
        self.loop = asyncio.get_event_loop()
        self.number_of_packets = 0
        self.data = []
        self.background_loop_thread: Optional[Thread] = None
        self.usart = None
        self.encoder = None
        self.ser_com_type = ser_com_type
        if self.ser_com_type == SerialCommunicationType.DLE_ENCODING:
            self.encoder = DleEncoder()
            self.reception_buffer = None
            # Set to default value.
            self.dle_queue_len = 10
            self.dle_max_frame = 256
            self.dle_timeout = 0.01

    def __del__(self):
        self.close()

    def is_open(self) -> bool:
        return True

    @property
    def id(self) -> str:
        return self.cfg.com_if_id

    def set_fixed_frame_settings(self, serial_frame_size: int):
        self.serial_frame_size = serial_frame_size

    def set_dle_settings(self, dle_queue_len: int, dle_max_frame: int, dle_timeout: float):
        self.dle_queue_len = dle_queue_len
        self.dle_max_frame = dle_max_frame
        self.dle_timeout = dle_timeout

    def initialize(self, _=None):
        """
        Needs to be called by application code once for DLE mode!
        """
        if not self.loop.is_running():
            self.background_loop_thread = Thread(
                target=start_background_loop, args=(self.loop,), daemon=True
            )

    def open(self, _=None) -> None:
        assert self.background_loop_thread is not None
        self.background_loop_thread.start()
        try:
            self.usart = Usart(self.cfg.serial_port)
            asyncio.run_coroutine_threadsafe(self.usart.open(), self.loop).result()
        except NotImplementedError:
            _LOGGER.exception("QEMU_SERIAL Initialization error, file does not exist!")
            sys.exit()
        if self.ser_com_type == SerialCommunicationType.DLE_ENCODING:
            self.reception_buffer = deque(maxlen=self.dle_queue_len)
            asyncio.run_coroutine_threadsafe(self.start_dle_polling(), self.loop)

    def close(self, _=None) -> None:
        if self.loop.is_closed():
            return

        assert self.usart is not None
        self.loop.call_soon_threadsafe(self.usart.close)
        self.loop.call_soon_threadsafe(self.loop.stop)

        while self.loop.is_running():
            time.sleep(0.01)

        self.loop.close()
        self.background_loop_thread.join()

    async def send_data_async(self, data):
        await self.usart.write(data)
        self.usart.inject_timeout_error()

    def send(self, data: bytearray):
        if self.ser_com_type == SerialCommunicationType.DLE_ENCODING:
            assert self.encoder is not None
            data_encoded = self.encoder.encode(data)
        else:
            data_encoded = data
        self.send_data(data_encoded)

    def send_data(self, data: bytearray):
        asyncio.run_coroutine_threadsafe(self.send_data_async(data), self.loop).result()

    def receive(self, _) -> TelemetryListT:
        assert self.usart is not None
        packet_list = []

        if self.ser_com_type == SerialCommunicationType.DLE_ENCODING:
            while self.reception_buffer:
                data = self.reception_buffer.pop()
                dle_retval, decoded_packet, _ = self.encoder.decode(data)
                if dle_retval == DleErrorCodes.OK:
                    packet_list.append(decoded_packet)
                else:
                    _LOGGER.warning("DLE decoder error!")

        else:
            _LOGGER.warning("This communication type was not implemented yet!")
        return packet_list

    def data_available(self, timeout: float = 0, _=0) -> int:
        if self.ser_com_type == SerialCommunicationType.DLE_ENCODING:
            return self.data_available_dle(timeout=timeout)
        return 0

    def data_available_fixed_frame(self, timeout: float = 0) -> int:
        assert self.usart is not None
        elapsed_time = 0
        start_time = time.time()
        sleep_time = timeout / 3.0
        if timeout > 0:
            while elapsed_time < timeout:
                if self.usart.new_data_available():
                    return self.usart.get_data_in_waiting()

                time.sleep(sleep_time)
                elapsed_time = time.time() - start_time
            return 0
        if self.usart.new_data_available():
            return self.usart.get_data_in_waiting()
        return 0

    def data_available_dle(self, timeout: float = 0) -> int:
        elapsed_time = 0
        start_time = time.time()
        sleep_time = timeout / 3.0
        if timeout > 0:
            while elapsed_time < timeout:
                if self.reception_buffer:
                    return len(self.reception_buffer)
                time.sleep(sleep_time)
                elapsed_time = time.time() - start_time
            return 0
        if self.reception_buffer:
            return len(self.reception_buffer)
        return 0

    async def start_dle_polling(self):
        asyncio.create_task(self.poll_dle_packets())

    async def poll_dle_packets(self):
        assert self.usart is not None
        assert self.reception_buffer is not None
        while True:
            rcvd = await self.usart.read_async(1, timeout=None)

            data = bytearray()
            data.append(rcvd[0])

            if data[0] == STX_CHAR:
                data.extend(
                    await self.usart.read_until_async(
                        bytes([ETX_CHAR]), DLE_FRAME_LENGTH, self.cfg.serial_timeout
                    )
                )

                # check for success
                if data[-1] == ETX_CHAR:
                    self.reception_buffer.appendleft(data)
                    continue

            else:  # not a start byte: flush input buffer
                data.extend(self.usart.read(self.usart.get_data_in_waiting()))

            # handle erroneous data
            print(data)
            # It is assumed that all packets are DLE encoded, so throw it away for now.
            _LOGGER.info("Non DLE-Encoded data with length " + str(len(data) + 1) + " found..")


class QmpException(Exception):
    """An exception caused by the QML/QEMU_SERIAL as response to a failed command"""

    def __init__(self, ret, *args, **kwargs):
        Exception.__init__(self, f"QMP error: {ret}")
        self.ret = ret  # the 'return' structure provided by QEMU_SERIAL/QML


class QmpConnection:
    """A connection to a QEMU_SERIAL machine via QMP"""

    def __init__(self, addr=QEMU_ADDR_QMP):
        self.transport = None
        self.addr = addr
        self.dataq = asyncio.Queue()
        self.initq = asyncio.Queue()
        self.proto = None

    def _protocol(self):
        """The underlying transport protocol"""

        if self.proto is None:
            self.proto = QmpProtocol(self)

        return self.proto

    async def _wait_check_return(self):
        """
        Wait for the status return of a command and raise an exception if it
        indicates a failure
        """

        resp = await self.dataq.get()
        if resp["return"]:
            raise QmpException(resp["return"])

    async def open(self):
        """
        Open this connection. Connect to the machine ensure that the
        connection is ready to use after this call.
        """

        loop = asyncio.get_running_loop()
        await loop.create_unix_connection(self._protocol, self.addr)

        # wait for initial capabilities and version
        init = await self.initq.get()
        print(init)

        # negotioate capabilities
        cmd = '{ "execute": "qmp_capabilities" }'
        self.transport.write(bytes(cmd, "utf-8"))
        await self._wait_check_return()

        return self

    def close(self):
        """Close this connection"""

        if self.transport is not None:
            self.transport.close()
            self.transport = None
            self.proto = None

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        self.close()

    async def cont(self):
        """Continue machine execution if it has been paused"""

        cmd = '{ "execute": "cont" }'
        self.transport.write(bytes(cmd, "utf-8"))
        await self._wait_check_return()

    async def stop(self):
        """Stop/pause machine execution"""

        cmd = '{ "execute": "stop" }'
        self.transport.write(bytes(cmd, "utf-8"))
        await self._wait_check_return()

    async def quit(self):
        """
        Quit the emulation. This causes the emulator to (non-gracefully)
        shut down and close.
        """

        cmd = '{ "execute": "quit" }'
        self.transport.write(bytes(cmd, "utf-8"))
        await self._wait_check_return()


class QmpProtocol(asyncio.Protocol):
    """The QMP transport protocoll implementation"""

    def __init__(self, conn):
        self.conn = conn

    def connection_made(self, transport):
        self.conn.transport = transport

    def connection_lost(self, exc):
        self.conn.transport = None
        self.conn.proto = None

    def data_received(self, data):
        data = str(data, "utf-8")
        decoder = json.JSONDecoder()
        nows = re.compile(r"[^\s]")

        pos = 0
        while True:
            match = nows.search(data, pos)
            if not match:
                return

            pos = match.start()
            obj, pos = decoder.raw_decode(data, pos)

            if "return" in obj:
                self.conn.dataq.put_nowait(obj)
            elif "QMP" in obj:
                self.conn.initq.put_nowait(obj)
            elif "event" in obj:
                pass
            else:
                print("qmp:", obj)


class DataFrame:
    """
    Basic protocol unit for communication via the IOX API introduced for
    external device emulation
    """

    def __init__(self, seq, cat, frame_id, data=None):
        self.seq = seq
        self.cat = cat
        self.id = frame_id
        self.data = data

    def bytes(self):
        """Convert this protocol unit to raw bytes"""
        data = self.data if self.data is not None else []
        return bytes([self.seq, self.cat, self.id, len(data)]) + bytes(data)

    def __repr__(self):
        return (
            f"{{ seq: 0x{self.seq:02x}, cat: 0x{self.cat:02x},"
            f" id: 0x{self.id:02x}, data: {self.data} }}"
        )


def parse_dataframes(buf):
    """Parse a variable number of DataFrames from the given byte buffer"""

    while len(buf) >= 4 and len(buf) >= 4 + buf[3]:
        frame = DataFrame(buf[0], buf[1], buf[2], buf[4 : 4 + buf[3]])
        buf = buf[4 + buf[3] :]
        yield buf, frame

    return buf, None


class UsartStatusException(Exception):
    """An exception returned by the USART send command"""

    def __init__(self, errn, *args, **kwargs):
        Exception.__init__(self, f"USART error: {errno.errorcode[errn]}")
        self.errno = errn  # a UNIX error code indicating the reason


class Usart:
    @staticmethod
    async def create_async(addr):
        return Usart(addr)

    """Connection to emulate a USART device for a given QEMU_SERIAL/At91 instance"""

    def __init__(self, addr):
        self.addr = addr
        self.respd = dict()
        self.respc = asyncio.Condition()
        self.dataq = asyncio.Queue()
        self.datab = bytes()
        self.transport = None
        self.proto = None
        self.seq = 0

    def _protocol(self):
        """The underlying transport protocol"""

        if self.proto is None:
            self.proto = UsartProtocol(self)

        return self.proto

    async def open(self):
        """Open this connection"""

        loop = asyncio.get_running_loop()
        await loop.create_unix_connection(self._protocol, self.addr)
        return self

    def close(self):
        """Close this connection"""

        if self.transport is not None:
            self.transport.close()
            self.transport = None
            self.proto = None

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        self.close()

    def _send_new_frame(self, cat, cid, data=None):
        """
        Send a DataFrame with the given parameters and auto-increase the
        sequence counter. Return its sequence number.
        """
        self.seq = (self.seq + 1) & 0x7F

        frame = DataFrame(self.seq, cat, cid, data)
        self.transport.write(frame.bytes())

        return frame.seq

    async def write(self, data):
        """Write data (bytes) to the USART device"""

        seq = self._send_new_frame(IOX_CAT_DATA, IOX_CID_DATA_IN, data)

        async with self.respc:
            while seq not in self.respd.keys():
                await self.respc.wait()

            resp = self.respd[seq]
            del self.respd[seq]

        status = struct.unpack("I", resp.data)[0]
        if status != 0:
            raise UsartStatusException(status)

    async def __read_async(self, n):
        while len(self.datab) < n:
            frame = await self.dataq.get()
            self.datab += frame.data

    async def read_async(self, n, timeout=None):
        """
        Wait for 'n' bytes to be received from the USART.

        This function will return early if the specified timeout (in
        seconds) is exceeded. In this case, only the data received up to
        that point will be returned. If timeout is None, no timeout will be
        set.
        """

        try:
            await asyncio.wait_for(self.__read_async(n), timeout)
        except asyncio.TimeoutError:
            pass  # ignore timeouts, return data received up to now
        finally:
            m = min(len(self.datab), n)
            data, self.datab = self.datab[:m], self.datab[m:]
            return data

    async def __read_until_async(self, expected, n):
        while n is None or len(self.datab) < n:
            frame = await self.dataq.get()
            self.datab += frame.data

            if expected in frame.data:
                return

    async def read_until_async(self, expected, size=None, timeout=None):
        """
        Read data until either the expected byte sequence has been found,
        the specified number of bytes has been received, or the timeout has
        occured.

        This function will return whatever data has been received up until
        the first termination condition has been met. In case size is None,
        there will be no size limit. In case timeout is None, ther will be
        no timeout.
        """
        try:
            await asyncio.wait_for(self.__read_until_async(expected, size), timeout)
        except asyncio.TimeoutError:
            pass  # ignore timeouts, return data received up to now
        finally:
            end = self.datab.find(expected)
            if end == -1:
                end = len(self.datab)
            else:
                end += len(expected)

            n = min(end, size)
            data, self.datab = self.datab[:n], self.datab[n:]
            return data

    def read(self, n):
        """Wait for 'n' bytes to be received from the USART
        timeout in seconds"""

        try:
            while len(self.datab) < n:
                frame = self.dataq.get_nowait()
                self.datab += frame.data
        # todo better solution
        finally:
            data, self.datab = self.datab[:n], self.datab[n:]
            return data

    def new_data_available(self) -> bool:
        return not self.dataq.empty()

    def get_data_in_waiting(self) -> int:
        return self.dataq.qsize()

    def flush(self):
        while True:
            try:
                self.dataq.get_nowait()
            except Exception as error:
                print(error)
                return

    def inject_overrun_error(self):
        """Inject an overrun error (set CSR_OVRE)"""
        self._send_new_frame(IOX_CAT_FAULT, IOX_CID_FAULT_OVRE)

    def inject_frame_error(self):
        """Inject a frame error (set CSR_FRAME)"""
        self._send_new_frame(IOX_CAT_FAULT, IOX_CID_FAULT_FRAME)

    def inject_parity_error(self):
        """Inject a parity error (set CSR_PARE)"""
        self._send_new_frame(IOX_CAT_FAULT, IOX_CID_FAULT_PARE)

    def inject_timeout_error(self):
        """Inject a timeout (set CSR_TIMEOUT)"""
        self._send_new_frame(IOX_CAT_FAULT, IOX_CID_FAULT_TIMEOUT)


class UsartProtocol(asyncio.Protocol):
    """The USART transport protocoll implementation"""

    def __init__(self, conn):
        self.conn = conn
        self.buf = bytes()

    def connection_made(self, transport):
        self.conn.transport = transport

    def connection_lost(self, exc):
        self.conn.transport = None
        self.conn.proto = None

    def data_received(self, data):
        self.buf += data

        for buf, frame in parse_dataframes(self.buf):
            self.buf = buf

            if frame.cat == IOX_CAT_DATA and frame.id == IOX_CID_DATA_OUT:
                # data from CPU/board to device
                self.conn.dataq.put_nowait(frame)
            elif frame.cat == IOX_CAT_DATA and frame.id == IOX_CID_DATA_IN:
                # response for data from device to CPU/board
                loop = asyncio.get_running_loop()
                loop.create_task(self._data_response_received(frame))

    async def _data_response_received(self, frame):
        async with self.conn.respc:
            self.conn.respd[frame.seq] = frame
            self.conn.respc.notify_all()



import os
import sys
import time
import unittest
from typing import Optional
from unittest import TestCase

from tmtccmd.com.serial_base import SerialCfg
from tmtccmd.com.serial_dle import SerialDleComIF


@unittest.skipIf(sys.platform.startswith("win"), "pty only works on POSIX systems")
class TestSerialDleInterface(TestCase):
    _DLE_IF: Optional[SerialDleComIF] = None
    _PTY_MASTER: Optional[int] = None

    # The ComIF will spawn a separate receiver thread. Therefore, I'd like to have one
    # instance for all the tests. This is done by using the class setup and teardown methods.
    @classmethod
    def setUpClass(cls) -> None:
        import pty

        cls._PTY_MASTER, slave = pty.openpty()
        sname = os.ttyname(slave)
        ser_cfg = SerialCfg(
            com_if_id="pseudo_ser_dle",
            serial_port=sname,
            baud_rate=9600,
            serial_timeout=1.0,
        )
        cls._DLE_IF = SerialDleComIF(ser_cfg, None)
        cls._DLE_IF.open()
        cls._DLE_IF.initialize()

    def setUp(self) -> None:
        from dle_encoder.dle_encoder import DleEncoder

        self.encoder = DleEncoder()

    def test_state(self):
        self.assertTrue(self._DLE_IF.is_open())
        self.assertEqual(self._DLE_IF.data_available(0), 0)
        self.assertEqual(self._DLE_IF.id, "pseudo_ser_dle")

    def test_send(self):
        test_data = bytes([0x01, 0x02, 0x03])
        encoded_len = len(self.encoder.encode(test_data))
        self._DLE_IF.send(test_data)
        encoded_packet = os.read(self._PTY_MASTER, encoded_len)
        (errors, test_data_read_back, len_read) = self.encoder.decode(encoded_packet)
        self.assertEqual(test_data_read_back, test_data)

    def test_recv(self):
        test_data = bytes([0x02, 0x03, 0x04])
        encoded_test_data = self.encoder.encode(test_data)
        os.write(self._PTY_MASTER, encoded_test_data)
        # Give the receiver thread some time to do its work.
        time.sleep(0.1)
        self.assertEqual(self._DLE_IF.data_available(0), 1)
        packet_list = self._DLE_IF.receive()
        self.assertEqual(len(packet_list), 1)
        # Received data should be decoded now
        self.assertEqual(packet_list[0], test_data)

    @classmethod
    def tearDownClass(cls) -> None:
        cls._DLE_IF.close()



"""PyQt front end components for the tmtccmd framework.
@author         R. Mueller, P. Scheurenbrand, D. Nguyen
"""

import os
from collections import deque
import sys
import webbrowser
from multiprocessing import Process
from pathlib import Path
from typing import Any, Tuple

from PyQt6.QtWidgets import (
    QLineEdit,
    QMainWindow,
    QGridLayout,
    QWidget,
    QLabel,
    QCheckBox,
    QDoubleSpinBox,
    QFrame,
    QComboBox,
    QPushButton,
    QMenu,
)
from PyQt6.QtGui import QPixmap, QIcon, QFont, QAction
from PyQt6.QtCore import (
    QTimer,
    Qt,
    QThreadPool,
)

from tmtccmd.core.base import FrontendBase
from tmtccmd.core.ccsds_backend import CcsdsTmtcBackend
from tmtccmd.config import HookBase
from tmtccmd.gui.buttons import (
    ConnectButtonParams,
    ButtonArgs,
    SendButtonWrapper,
    TmButtonWrapper,
    ConnectButtonWrapper,
)
from tmtccmd.gui.cmd_select import CommandPathSelectWidget
from tmtccmd.gui.defs import SharedArgs, CONNECT_BTTN_STYLE, FrontendState
from tmtccmd.logging import get_console_logger
import tmtccmd as mod_root

LOGO_PATH = Path(f"{Path(mod_root.__file__).parent.parent}/misc/logo-tiny.png")


LOGGER = get_console_logger()


class TmTcFrontend(QMainWindow, FrontendBase):
    def __init__(self, hook_obj: HookBase, tmtc_backend: CcsdsTmtcBackend, app_name: str):
        super(TmTcFrontend, self).__init__()
        super(QMainWindow, self).__init__()
        self._app_name = app_name
        self._shared_args = SharedArgs(tmtc_backend)
        tmtc_backend.exit_on_com_if_init_failure = False
        self._hook_obj = hook_obj
        self._com_if_list = []
        self._last_selected_items = deque()
        self._state = FrontendState()
        self._thread_pool = QThreadPool()
        self.logo_path = LOGO_PATH

    def prepare_start(self, _: Any) -> Process:
        return Process(target=self.start)

    def start(self, qt_app: Any):
        self._start_ui()
        self._enable_periodic_interpreter_run()
        sys.exit(qt_app.exec())

    def _enable_periodic_interpreter_run(self):
        self._interpreter_run_timer = QTimer()
        self._interpreter_run_timer.start(500)  # You may change this if you wish.
        self._interpreter_run_timer.timeout.connect(
            lambda: None
        )  # Let the interpreter run each 500 ms.

    def _start_ui(self):
        self._create_menu_bar()
        win = QWidget(self)
        self.setCentralWidget(win)

        grid = QGridLayout()
        win.setLayout(grid)
        row = 0
        self.setWindowTitle(self._app_name)
        if isinstance(self.logo_path, Path):
            self.setWindowIcon(QIcon(str(self.logo_path)))
        else:
            self.setWindowIcon(QIcon(self.logo_path))

        add_pixmap = False

        if add_pixmap:
            row = self.__set_up_pixmap(grid=grid, row=row)

        row = self.__set_up_config_section(grid=grid, row=row)
        row = self.__add_vertical_separator(grid=grid, row=row)

        tm_listener_button = QPushButton()
        conn_bttn_params = ConnectButtonParams(
            hook_obj=self._hook_obj,
            connect_cb=self.__connected_com_if_cb,
            disconnect_cb=self.__disconnect_com_if_cb,
            tm_listener_bttn=tm_listener_button,
        )
        # com if configuration
        row, self.__connect_button_wrapper = self.__set_up_com_if_section(
            conn_bttn_params=conn_bttn_params, grid=grid, row=row
        )
        row = self.__add_vertical_separator(grid=grid, row=row)

        tmtc_ctrl_label = QLabel("TMTC Control")
        font = QFont()
        font.setBold(True)
        tmtc_ctrl_label.setFont(font)
        grid.addWidget(tmtc_ctrl_label, row, 0, 1, 2)
        row += 1
        row = self._set_up_cmd_path_ui(grid=grid, row=row)

        button_args = ButtonArgs(
            state=self._state, pool=self._thread_pool, shared=self._shared_args
        )
        self.__send_bttn_wrapper = SendButtonWrapper(
            button=QPushButton(),
            args=button_args,
            conn_button=self.__connect_button_wrapper.button,
        )
        grid.addWidget(self.__send_bttn_wrapper.button, row, 0, 1, 2)
        row += 1

        self.__tm_button_wrapper = TmButtonWrapper(
            button=tm_listener_button,
            args=button_args,
            conn_button=self.__connect_button_wrapper.button,
        )
        grid.addWidget(self.__tm_button_wrapper.button, row, 0, 1, 2)
        row += 1
        self.show()
        # self.destroyed.connect(self.__tm_button_wrapper.stop_thread)

    def set_gui_logo(self, logo_total_path: str):
        if os.path.isfile(logo_total_path):
            self.logo_path = logo_total_path
        else:
            LOGGER.warning("Could not set logo, path invalid!")

    def closeEvent(self, event):
        try:
            pass
            if self.__tm_button_wrapper.is_listening():
                LOGGER.warning("TM listener still active. Stopping it first..")
                self.__tm_button_wrapper.stop_thread()
                event.ignore()
            else:
                pass
        except KeyboardInterrupt:
            self.__tm_button_wrapper.abort_thread()

    def _create_menu_bar(self):
        menu_bar = self.menuBar()
        # Creating menus using a QMenu object
        file_menu = QMenu("&File", self)
        assert menu_bar is not None
        menu_bar.addMenu(file_menu)
        # Creating menus using a title
        help_menu = menu_bar.addMenu("&Help")
        assert help_menu is not None

        help_action = QAction("Help", self)
        help_action.triggered.connect(self.__help_url)
        help_menu.addAction(help_action)

    @staticmethod
    def __help_url():
        webbrowser.open("https://tmtccmd.readthedocs.io/en/latest/")

    def __set_up_config_section(self, grid: QGridLayout, row: int) -> int:
        label = QLabel("Configuration")
        font = QFont()
        font.setBold(True)
        label.setFont(font)
        grid.addWidget(label, row, 0, 1, 2)
        row += 1

        start_listener_on_connect = QCheckBox("Auto-Connect TM listener")
        start_listener_on_connect.setChecked(True)
        start_listener_on_connect.stateChanged.connect(
            lambda: self._tm_auto_connect_changed(start_listener_on_connect)
        )
        grid.addWidget(start_listener_on_connect, row, 0, 1, 1)
        row += 1

        grid.addWidget(QLabel("Inter-Packet Delay Seconds [0 - 500]"), row, 0, 1, 2)
        row += 1

        spin_timeout = QDoubleSpinBox()
        spin_timeout.setValue(0.1)
        # TODO: set sensible min/max values
        spin_timeout.setSingleStep(0.1)
        spin_timeout.setMinimum(0.0)
        spin_timeout.setMaximum(500.0)
        # https://youtrack.jetbrains.com/issue/PY-22908
        # Ignore those warnings for now.
        spin_timeout.valueChanged.connect(self.__number_timeout_changed)
        grid.addWidget(spin_timeout, row, 0, 1, 1)
        row += 1
        return row

    def _tm_auto_connect_changed(self, box: QCheckBox):
        if box.isChecked():
            self._state.auto_connect_tm_listener = True
        else:
            self._state.auto_connect_tm_listener = False

    def __set_up_com_if_section(
        self, conn_bttn_params: ConnectButtonParams, grid: QGridLayout, row: int
    ) -> Tuple[int, ConnectButtonWrapper]:
        font = QFont()
        font.setBold(True)
        label = QLabel("Communication Interface")
        label.setFont(font)
        grid.addWidget(label, row, 0, 1, 1)
        com_if_combo_box = QComboBox()
        all_com_ifs = self._hook_obj.get_com_if_dict()
        index = 0
        # add all possible ComIFs to the comboBox
        for id, com_if_value in all_com_ifs.items():
            com_if_combo_box.addItem(com_if_value[0])
            self._com_if_list.append((id, com_if_value[0]))
            if self._shared_args.backend.com_if_id == id:
                com_if_combo_box.setCurrentIndex(index)
                self._state.current_com_if = id
            index += 1
        com_if_combo_box.currentIndexChanged.connect(self.__com_if_sel_index_changed)
        grid.addWidget(com_if_combo_box, row, 1, 1, 1)
        row += 1

        self.com_if_cfg_button = QPushButton()
        self.com_if_cfg_button.setText("Configure")
        grid.addWidget(self.com_if_cfg_button, row, 0, 1, 2)
        row += 1

        connect_button = QPushButton()
        connect_button.setText("Connect")
        connect_button.setStyleSheet(CONNECT_BTTN_STYLE)
        conn_bttn_wrapper = ConnectButtonWrapper(
            button=connect_button,
            args=ButtonArgs(self._state, self._thread_pool, self._shared_args),
            bttn_params=conn_bttn_params,
        )
        grid.addWidget(connect_button, row, 0, 1, 2)
        row += 1
        return row, conn_bttn_wrapper

    def __disable_conn_bttn(self):
        self.__connect_button_wrapper.button.setDisabled(True)

    def __enable_conn_bttn(self):
        self.__connect_button_wrapper.button.setEnabled(True)

    def __connected_com_if_cb(self):
        self.__send_bttn_wrapper.button.setEnabled(True)
        self.__tm_button_wrapper.button.setEnabled(True)

    def __disconnect_com_if_cb(self):
        self.__send_bttn_wrapper.button.setDisabled(True)
        self.__tm_button_wrapper.button.setDisabled(True)

    def _set_up_cmd_path_ui(self, grid: QGridLayout, row: int):
        font = QFont()
        font.setBold(True)
        label = QLabel("Communication Select")
        label.setFont(font)
        grid.addWidget(label, row, 0, 1, 1)
        row += 1
        self._open_command_path_select_button = QPushButton("Command Path Select")
        self._open_command_path_select_button.clicked.connect(self._open_command_select_widget)
        grid.addWidget(self._open_command_path_select_button, row, 0, 1, 2)
        row += 1
        self._cmd_path_text_input = QLineEdit()
        grid.addWidget(self._cmd_path_text_input, row, 0, 1, 2)
        row += 1
        self._confirm_selected_cmd_path_button = QPushButton("Confirm selected path")
        grid.addWidget(self._confirm_selected_cmd_path_button, row, 0, 1, 2)
        row += 1
        self._cmd_path_text = QLabel("Selected command path: ")
        grid.addWidget(self._cmd_path_text, row, 0, 1, 2)
        row += 1
        return row

    def _open_command_select_widget(self):
        self.cmd_select_window = CommandPathSelectWidget(
            self._hook_obj.get_command_definitions(), self._last_selected_items
        )
        self.cmd_select_window.path_selected_sig.connect(self._receive_selected_path)
        self.cmd_select_window.closed.connect(self._on_treeview_closed)
        self.cmd_select_window.show()
        self._open_command_path_select_button.setEnabled(False)

    def _set_cmd_path_label(self, text: str):
        self._cmd_path_text.setText(f"Selected command path: {text}")

    def _confirm_selected_cmd_path(self):
        self._set_cmd_path_label(self._cmd_path_text_input.text())
        self._state.current_cmd_path = self._cmd_path_text_input.text()

    def _on_treeview_closed(self):
        self._open_command_path_select_button.setEnabled(True)

    def _receive_selected_path(self, path: str):
        self._cmd_path_text_input.setText(path)
        self._confirm_selected_cmd_path()

    def __set_up_pixmap(self, grid: QGridLayout, row: int) -> int:
        label = QLabel(self)
        label.setGeometry(720, 10, 100, 100)
        label.adjustSize()

        pixmap = QPixmap(self.logo_path)
        pixmap_width = pixmap.width()
        pixmap_height = pixmap.height()
        row += 1

        pixmap_scaled = pixmap.scaled(
            int(pixmap_width * 0.3),
            int(pixmap_height * 0.3),
            Qt.AspectRatioMode.KeepAspectRatio,
        )
        label.setPixmap(pixmap_scaled)
        label.setScaledContents(True)

        grid.addWidget(label, row, 0, 1, 2)
        row += 1
        return row

    @staticmethod
    def __add_vertical_separator(grid: QGridLayout, row: int):
        separator = QFrame()
        separator.setFrameShape(QFrame.Shape.HLine)
        grid.addWidget(separator, row, 0, 1, 2)
        row += 1
        return row

    def __com_if_sel_index_changed(self, index: int):
        self._state.current_com_if = self._com_if_list[index][0]
        LOGGER.debug(f"Communication IF updated: {self._com_if_list[index][1]}")

    def __number_timeout_changed(self, value: float):
        LOGGER.info("PUS TM timeout changed to: " + str(value))



"""Contains definitions and functions related to PUS Service 5 Telecommands."""

from deprecated.sphinx import deprecated
from spacepackets.ecss import PusTelecommand
from spacepackets.ecss.pus_5_event import Subservice


@deprecated(
    version="v4.0.0a0",
    reason="use create... API instead",
)
def pack_enable_event_reporting_command(apid: int = 0, seq_count: int = 0) -> PusTelecommand:
    return create_disable_event_reporting_command(apid, seq_count)


def create_enable_event_reporting_command(apid: int = 0, seq_count: int = 0) -> PusTelecommand:
    return PusTelecommand(
        service=5,
        subservice=Subservice.TC_ENABLE_EVENT_REPORTING,
        seq_count=seq_count,
        apid=apid,
    )


@deprecated(
    version="v4.0.0a0",
    reason="use create... API instead",
)
def pack_disable_event_reporting_command(apid: int = 0, seq_count: int = 0) -> PusTelecommand:
    return create_disable_event_reporting_command(apid, seq_count)


def create_disable_event_reporting_command(apid: int = 0, seq_count: int = 0) -> PusTelecommand:
    return PusTelecommand(
        service=5,
        subservice=Subservice.TC_DISABLE_EVENT_REPORTING,
        seq_count=seq_count,
        apid=apid,
    )



from pathlib import Path
from unittest import TestCase

from spacepackets.cfdp import CfdpLv
from spacepackets.cfdp.tlv import ProxyMessageType
from spacepackets.util import ByteFieldU8
from cfdppy.request import PutRequest
from tmtccmd.config.cfdp import (
    cfdp_req_to_put_req_regular,
    cfdp_req_to_put_req_get_req,
    generic_cfdp_params_to_put_request,
    CfdpParams,
)


class TestCfdpParamsConversion(TestCase):
    def setUp(self) -> None:
        self.test_id_local = ByteFieldU8(1)
        self.test_dest_id_remote = ByteFieldU8(2)

    def _verify_regular_put_request(self, cfdp_params: CfdpParams, put_request: PutRequest):
        self.assertEqual(put_request.source_file, Path(cfdp_params.source_file))
        self.assertEqual(put_request.dest_file, Path(cfdp_params.dest_file))

    def test_put_request_generation(self):
        cfdp_params = CfdpParams(source_file="hello.txt", dest_file="hello2.txt")
        put_request = cfdp_req_to_put_req_regular(cfdp_params, self.test_dest_id_remote)
        self._verify_regular_put_request(cfdp_params, put_request)

    def test_generic_put_request_generation(self):
        cfdp_params = CfdpParams(source_file="hello.txt", dest_file="hello2.txt")
        put_request = generic_cfdp_params_to_put_request(
            cfdp_params,
            local_id=self.test_id_local,
            remote_id=self.test_dest_id_remote,
            dest_id_proxy_put_req=self.test_id_local,
        )
        self._verify_regular_put_request(cfdp_params, put_request)

    def test_invalid_conversion_for_proxy_op(self):
        cfdp_param = CfdpParams(proxy_op=True)
        self.assertIsNone(cfdp_req_to_put_req_regular(cfdp_param, self.test_dest_id_remote))

    def _verify_get_request(self, put_request: PutRequest, cfdp_params: CfdpParams):
        self.assertEqual(put_request.destination_id, self.test_dest_id_remote)
        self.assertIsNotNone(put_request.msgs_to_user)
        self.assertEqual(len(put_request.msgs_to_user), 1)
        msg_to_user = put_request.msgs_to_user[0]
        self.assertTrue(msg_to_user.is_reserved_cfdp_message())
        reserved_msg = msg_to_user.to_reserved_msg_tlv()
        self.assertTrue(reserved_msg.is_cfdp_proxy_operation())
        self.assertEqual(reserved_msg.get_cfdp_proxy_message_type(), ProxyMessageType.PUT_REQUEST)
        proxy_put_req_params = reserved_msg.get_proxy_put_request_params()
        self.assertIsNotNone(proxy_put_req_params)
        self.assertEqual(proxy_put_req_params.dest_entity_id, self.test_id_local)
        self.assertEqual(
            proxy_put_req_params.source_file_name,
            CfdpLv.from_str(cfdp_params.source_file),
        )
        self.assertEqual(
            proxy_put_req_params.dest_file_name, CfdpLv.from_str(cfdp_params.dest_file)
        )

    def test_get_request_put_request_generation(self):
        cfdp_params = CfdpParams(proxy_op=True, source_file="hello.txt", dest_file="hello2.txt")
        put_request = cfdp_req_to_put_req_get_req(
            cfdp_params, self.test_id_local, self.test_dest_id_remote
        )
        self._verify_get_request(put_request, cfdp_params)

    def test_generic_conversion_function(self):
        cfdp_params = CfdpParams(proxy_op=True, source_file="hello.txt", dest_file="hello2.txt")
        put_request = generic_cfdp_params_to_put_request(
            cfdp_params,
            self.test_id_local,
            self.test_dest_id_remote,
            self.test_id_local,
        )
        self._verify_get_request(put_request, cfdp_params)



"""`FSFW <https://egit.irs.uni-stuttgart.de/fsfw/fsfw>`_ specific PUS actions support"""

from .s8_fsfw_action_defs import *  # noqa re-export
from .tc.s8_fsfw_action import *  # noqa re-export
from .tm.s8_fsfw_action import *  # noqa re-export



"""TCP communication interface"""

import logging
import queue
import socket
import time
import enum
import threading
import select
from collections import deque
from typing import Any, List, Optional, Sequence

from spacepackets.ccsds.spacepacket import parse_space_packets, PacketId

from tmtccmd.com import ComInterface, SendError
from tmtccmd.com.tcpip_utils import EthAddr

_LOGGER = logging.getLogger(__name__)

TCP_RECV_WIRETAPPING_ENABLED = False
TCP_SEND_WIRETAPPING_ENABLED = False


class TcpCommunicationType(enum.Enum):
    """Parse for space packets in the TCP stream, using the space packet header."""

    SPACE_PACKETS = 0


class TcpSpacepacketsClient(ComInterface):
    """Communication interface for TCP communication. This particular interface expects
    raw space packets to be sent via TCP and uses a list of passed packet IDs to parse for them.
    """

    def __init__(
        self,
        com_if_id: str,
        space_packet_ids: Sequence[PacketId],
        inner_thread_delay: float,
        target_address: EthAddr,
        max_packets_stored: Optional[int] = None,
    ):
        """Initialize a communication interface to send and receive TMTC via TCP.

        :param com_if_id:
        :param space_packet_ids: Valid packet IDs for CCSDS space packets. Those will be used
            to parse for space packets inside the TCP stream.
        :param inner_thread_delay: Polling frequency of TCP thread in seconds.
        """
        self.com_if_id = com_if_id
        self.com_type = TcpCommunicationType.SPACE_PACKETS
        self.space_packet_ids = space_packet_ids
        self.__inner_thread_delay = inner_thread_delay
        self.target_address = target_address
        self.max_packets_stored = max_packets_stored
        self.__conn_lock = threading.Lock()
        self.__connected = False
        self.__tcp_socket = None
        self.__thread_kill_signal = threading.Event()
        # Separate thread to request TM packets periodically if no TCs are being sent
        self.__tcp_thread = None
        self.__tm_queue = queue.Queue()
        self.__tc_queue = queue.Queue()
        self.__analysis_queue = deque()
        self.tm_packet_list = []

    @property
    def id(self) -> str:
        return self.com_if_id

    def __del__(self):
        try:
            self.close()
        except IOError:
            _LOGGER.warning("Could not close TCP communication interface!")

    def initialize(self, args: Any = None):
        pass

    def open(self, args: Any = None):
        if self.is_open():
            return
        self.__thread_kill_signal.clear()
        try:
            self.__init_socket()
            self.__connect_socket()
        except IOError as e:
            _LOGGER.exception("Issues setting up the TCP socket")
            raise e
        if self.__tcp_thread is None:
            self.__tcp_thread = threading.Thread(target=self.__tcp_task)
            self.__tcp_thread.start()
            with self.__conn_lock:
                self.__connected = True

    def is_open(self) -> bool:
        with self.__conn_lock:
            return self.__connected

    def __init_socket(self):
        if self.__tcp_socket is None:
            self.__tcp_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.__tcp_socket.settimeout(2.0)

    def __connect_socket(self):
        assert self.__tcp_socket is not None
        try:
            self.__tcp_socket.connect(self.target_address.to_tuple)
        except socket.timeout as e:
            _LOGGER.warning(
                "Could not connect to socket with address" f" {self.target_address}: {e}"
            )
        finally:
            self.__tcp_socket.settimeout(None)

    def close(self, args: any = None) -> None:
        if not self.is_open():
            return
        self.__thread_kill_signal.set()
        if self.__tcp_thread is not None:
            self.__tcp_thread.join(self.__inner_thread_delay)
            with self.__conn_lock:
                self.__connected = False
        self.__tcp_socket = None

    def send(self, data: bytes):
        self.__tc_queue.put(data)

    def receive(self, poll_timeout: float = 0) -> List[bytes]:
        self.__tm_queue_to_packet_list()
        tm_packet_list = self.tm_packet_list
        self.tm_packet_list = []
        return tm_packet_list

    def __tm_queue_to_packet_list(self):
        while self.__tm_queue.qsize() > 0:
            self.__analysis_queue.append(self.__tm_queue.get())
        # TCP is stream based, so there might be broken packets or multiple packets in one recv
        # call. We parse the space packets contained in the stream here
        if self.com_type == TcpCommunicationType.SPACE_PACKETS:
            self.tm_packet_list.extend(
                parse_space_packets(
                    analysis_queue=self.__analysis_queue,
                    packet_ids=self.space_packet_ids,
                )
            )
        else:
            while self.__analysis_queue:
                self.tm_packet_list.append(self.__analysis_queue.popleft())

    def __tcp_task(self):
        while True and not self.__thread_kill_signal.is_set():
            try:
                self.__tmtc_event_loop()
            except ConnectionRefusedError:
                _LOGGER.warning("TCP connection attempt failed..")
            time.sleep(self.__inner_thread_delay)

    def __tmtc_event_loop(self):
        assert self.__tcp_socket is not None
        try:
            while True:
                outputs = []
                queue_size = self.__tc_queue.qsize()
                if queue_size > 0:
                    outputs.append(self.__tcp_socket)
                (readable, writable, _) = select.select(
                    [self.__tcp_socket], outputs, [], self.__inner_thread_delay
                )
                if self.__thread_kill_signal.is_set():
                    self.__tcp_socket.close()
                    break
                if queue_size > 0 and writable and writable[0]:
                    self.__tc_handling(queue_size)
                if readable and readable[0]:
                    self.__tm_handling()
        except KeyboardInterrupt:
            _LOGGER.info("Keyboard interrupt, shutting down TCP task")
            self.__force_shutdown()
        except ConnectionResetError:
            self.__force_shutdown()
            _LOGGER.exception("ConnectionResetError. TCP server might not be up")

    def __tc_handling(self, queue_size: int):
        try:
            self.__tcp_socket.sendto(self.__tc_queue.get(), self.target_address.to_tuple)
            queue_size -= 1
        except BrokenPipeError as e:
            raise SendError(f"{e}", e)
        except ConnectionRefusedError or OSError as e:
            self.__force_shutdown()
            raise SendError(f"TCP connection attempt failed with exception: {e}", e)

    def __tm_handling(self):
        bytes_recvd = self.__tcp_socket.recv(4096)
        if bytes_recvd == b"":
            self.__force_shutdown()
            _LOGGER.info("TCP server has been closed")
            return
        if (
            self.max_packets_stored is not None
            and self.__tm_queue.qsize() >= self.max_packets_stored
        ):
            _LOGGER.warning(
                "Number of packets in TCP queue too large. " "Overwriting old packets.."
            )
            self.__tm_queue.get()
            # TODO: If segments are received but the receiver is unable to parse packets
            #       properly, it might make sense to have a timeout which then also
            #       logs that there might be an issue reading packets
        self.__tm_queue.put(bytes(bytes_recvd))

    def data_available(self, timeout: float = 0, parameters: any = 0) -> int:
        self.__tm_queue_to_packet_list()
        return len(self.tm_packet_list)

    def __force_shutdown(self):
        assert self.__tcp_socket is not None
        self.__tcp_socket.close()
        with self.__conn_lock:
            self.__connected = False



"""Argument parser module."""

from __future__ import annotations

import argparse
import logging
import os
import sys
from dataclasses import dataclass
from typing import List, Optional, Sequence, Tuple, Union

from prompt_toolkit.shortcuts import CompleteStyle
from spacepackets.cfdp import TransmissionMode

from tmtccmd.com import ComInterface
from tmtccmd.com.utils import determine_com_if
from tmtccmd.config.prompt import prompt_cmd_path
from tmtccmd.config.tmtc import CmdTreeNode
from tmtccmd.tmtc.procedure import TcProcedureType

from .defs import (
    CfdpParams,
    CoreComInterfaces,
    CoreModeConverter,
    CoreModeList,
    TreeCommandingParams,
)
from .hook import HookBase

_LOGGER = logging.getLogger(__name__)


def get_default_descript_txt() -> str:
    from tmtccmd.util.conf_util import AnsiColors

    return (
        f"{AnsiColors.GREEN}TMTC Client Command Line Interface\n{AnsiColors.RESET}This"
        " application provides generic components to perform remote\ncommanding with"
        " special support for space applications.\n"
    )


def create_default_args_parser(
    parent_parser: argparse.ArgumentParser,
    descript_txt: Optional[str] = None,
) -> argparse.ArgumentParser:
    if descript_txt is None:
        descript_txt = get_default_descript_txt()
    return argparse.ArgumentParser(
        description=descript_txt,
        formatter_class=argparse.RawTextHelpFormatter,
        parents=[parent_parser],
    )


class ProcedureParamsWrapper:
    def __init__(self):
        self._ptype = TcProcedureType.CUSTOM
        self._params = None

    @property
    def ptype(self):
        return self._ptype

    def set_params(self, params: Union[TreeCommandingParams, CfdpParams]):
        if isinstance(params, TreeCommandingParams):
            self._params = params
            self._ptype = TcProcedureType.TREE_COMMANDING
        elif isinstance(params, CfdpParams):
            self._params = params
            self._ptype = TcProcedureType.CFDP

    def tree_commanding_params(self) -> Optional[TreeCommandingParams]:
        if self._ptype == TcProcedureType.TREE_COMMANDING:
            return self._params
        return None

    def cfdp_params(self) -> Optional[CfdpParams]:
        if self._ptype == TcProcedureType.CFDP:
            return self._params
        return None


@dataclass
class CommandingParams:
    delay: float = 0.0
    apid: int = 0
    print_tree: bool = False
    tree_print_with_description: bool = True
    tree_print_max_depth: Optional[int] = None


@dataclass
class BackendParams:
    mode: str = ""
    com_if_id: str = ""
    listener: bool = False
    interactive: bool = False


@dataclass
class AppParams:
    use_gui: bool = False
    reduced_printout: bool = False
    use_ansi_colors: bool = True
    compl_style: CompleteStyle = CompleteStyle.READLINE_LIKE


class SetupParams:
    def __init__(
        self,
        com_if: Optional[ComInterface] = None,
        cmd_params: Optional[CommandingParams] = None,
        backend_params: Optional[BackendParams] = None,
        app_params: Optional[AppParams] = None,
    ):
        self.com_if = com_if
        if cmd_params is None:
            self.cmd_params = CommandingParams()
        if backend_params is None:
            self.backend_params = BackendParams()
        if app_params is None:
            self.app_params = AppParams()

    @property
    def apid(self):
        return self.cmd_params.apid

    @apid.setter
    def apid(self, apid):
        self.cmd_params.apid = apid

    @property
    def use_gui(self):
        return self.app_params.use_gui

    @use_gui.setter
    def use_gui(self, use_gui):
        self.app_params.use_gui = use_gui

    @property
    def mode(self):
        return self.backend_params.mode

    @mode.setter
    def mode(self, mode: str):
        self.backend_params.mode = mode

    @property
    def com_if_id(self):
        return self.backend_params.com_if_id

    @com_if_id.setter
    def com_if_id(self, com_if_id):
        self.backend_params.com_if_id = com_if_id


def add_default_tmtccmd_args(parser: argparse.ArgumentParser):
    add_tmtc_mode_arguments(parser)
    add_default_com_if_arguments(parser)
    add_generic_arguments(parser)
    add_tree_commanding_arguments(parser)
    add_ethernet_arguments(parser)


def parse_default_tmtccmd_input_arguments(
    args: Sequence[str],
    parser: argparse.ArgumentParser,
    print_known_args: bool = False,
    print_unknown_args: bool = False,
) -> Tuple[argparse.Namespace, List[str]]:
    """Parses all input arguments by calling :py:meth:`argparse.ArgumentParser.parse_known_args`.
    It is recommended to use the :py:class:`PreArgsParsingWrapper` instead of using this function
    directly.

    :param args: The actual full list of parse CLI arguments
    :param parser: The parser to be used.
    :param print_known_args: Debugging function to print all known arguments.
    :param print_unknown_args:
    :return: Input arguments contained in a special namespace and accessable by args.<variable>
    """

    if len(sys.argv) == 1:
        print("No input arguments specified. Run with -h to get list of arguments")

    args, unknown = parser.parse_known_args(args)

    if print_known_args:
        _LOGGER.info("Printing known arguments:")
        for argument in vars(args):
            _LOGGER.debug(argument + ": " + str(getattr(args, argument)))
    if print_unknown_args:
        _LOGGER.info("Printing unknown arguments:")
        for argument in unknown:
            _LOGGER.info(argument)

    if len(unknown) > 0:
        print(f"Unknown arguments detected: {unknown}")
    return args, unknown


def add_tree_commanding_arguments(parser_or_subparser: argparse.ArgumentParser):
    parser_or_subparser.add_argument(
        "-p",
        "--path",
        dest="cmd_path",
        help="Command tree path, used to uniquely identify command or command stack to be sent.",
        default=None,
    )
    parser_or_subparser.add_argument(
        "-T",
        "--pt",
        "--print-tree",
        dest="print_tree",
        nargs="*",
        default=None,
        help=(
            f"Optional arguments [b] [p] [<numMaxDepth>]. Print the command definition tree. You "
            f"can{os.linesep}optionally add b to omit descriptions, p to display hidden nodes, "
            f"and a maximum print depth."
        ),
    )
    add_tmtc_mode_arguments(parser_or_subparser)
    add_tmtc_listener_arg(parser_or_subparser)


def add_cfdp_procedure_arguments(parser_or_subparser: argparse.ArgumentParser):
    """TODO: Could be extended to support the various types of CFDP user primitives.
    Right now, the first thing to be implemented will be the put request"""
    parser_or_subparser.add_argument(
        "source", help="Full Source Path for File Copy Procedures", nargs="?"
    )
    parser_or_subparser.add_argument(
        "target", help="Full Destination Path for File Copy Procedures", nargs="?"
    )
    parser_or_subparser.add_argument(
        "-p",
        "--proxy",
        action="store_true",
        help=(
            f"Used to trigger a proxy operation at the remote CFDP entity.{os.linesep}Most"
            f" commonly used to request a file from the remote entity.{os.linesep}Please note that"
            f" this inverses the meaning of the destination and file parameter."
        ),
    )
    parser_or_subparser.add_argument(
        "-t",
        "--type",
        help=(
            f"Specify the transfer type{os.linesep}"
            f' - "0" or "ack" for unacknowledged (Class 0) transfers{os.linesep}'
            f' - "1" or "nak" for acknowledged (Class 1) transfers. Default value'
        ),
        default="nak",
    )
    parser_or_subparser.add_argument(
        "--no-closure",
        help="Disable the requesting of transaction closure",
        action="store_true",
        dest="no_closure",
    )


def add_generic_arguments(arg_parser: argparse.ArgumentParser):
    arg_parser.add_argument("-g", "--gui", help="Use GUI mode", action="store_true", default=False)
    arg_parser.add_argument(
        "-i",
        "--interactive",
        help=(
            "Enables interactive or multi-queue mode, where the backend will be"
            " configured\nto handle multiple queues"
        ),
        action="store_true",
        default=False,
    )
    arg_parser.add_argument(
        "-d",
        "--delay",
        type=float,
        help=(
            "Default inter-packet delay. Default:\n"
            " - Default One-Queue Mode: 4 seconds\n"
            " - Multi-Queue Mode: 0 seconds"
        ),
        default=None,
    )


def add_tmtc_mode_arguments(arg_parser: argparse.ArgumentParser):
    from tmtccmd.config import CoreModeConverter, CoreModeList

    help_text = (
        "Core Modes. Default:" f" {CoreModeConverter.get_str(CoreModeList.ONE_QUEUE_MODE)}\n"
    )
    one_q = (
        f' - "{CoreModeList.ONE_QUEUE_MODE}" or '
        f'"{CoreModeConverter.get_str(CoreModeList.ONE_QUEUE_MODE)}": '
        "One Queue Command Mode\n"
    )
    listener_help = (
        f' - "{CoreModeList.LISTENER_MODE}" or '
        f'"{CoreModeConverter.get_str(CoreModeList.LISTENER_MODE)}": Listener Mode\n'
    )
    multi_q = (
        f' - "{CoreModeList.MULTI_INTERACTIVE_QUEUE_MODE}" or '
        f'"{CoreModeConverter.get_str(CoreModeList.MULTI_INTERACTIVE_QUEUE_MODE)}": '
        "Multi Queue and Interactive Command Mode\n"
    )
    help_text += one_q + listener_help + multi_q
    arg_parser.add_argument(
        "-m",
        "--mode",
        type=str,
        help=help_text,
        default=None,
    )


def add_tmtc_listener_arg(arg_parser: argparse.ArgumentParser):
    arg_parser.add_argument(
        "-l",
        "--listener",
        help=(
            "The backend will be configured to go into listener mode after finishing"
            " the\nfirst queue if a service argument is specified. If this flag is"
            " specified\nwithout the -s flag and none of the queue modes are specified"
            " explicitly,\nthe mode will be set to the listener mode"
        ),
        action="store_true",
        default=False,
    )
    arg_parser.add_argument(
        "--pp",
        "--prompt-proc",
        help=(
            "If this is specified in addition to the listener mode flag -l, the"
            " commander will\nprompt for the next command or procedure to be sent."
        ),
        dest="prompt_proc",
        action="store_true",
        default=False,
    )


def add_default_com_if_arguments(arg_parser: argparse.ArgumentParser):
    from tmtccmd.config import CORE_COM_IF_DICT, CoreComInterfaces

    help_text = (
        "Core Communication Interface. If this is not specified, the commander core\n"
        "will try to extract it from the JSON or prompt it from the user.\n"
        "Choices provided by framework:\n"
    )
    for k, v in CORE_COM_IF_DICT.items():
        help_text += f' - "{k}": {v[0]}\n'
    arg_parser.add_argument(
        "-c",
        "--com_if",
        type=str,
        help=help_text,
        default=CoreComInterfaces.UNSPECIFIED.value,
    )


def add_ethernet_arguments(arg_parser: argparse.ArgumentParser):
    arg_parser.add_argument("--h-ip", help="Host (Computer) IP. Default:''", default="")
    arg_parser.add_argument(
        "--t-ip", help="Target IP. Default: Localhost 127.0.0.1", default="127.0.0.1"
    )


def determine_cmd_path(
    params: SetupParams,
    def_params: TreeCommandingParams,
    hook_obj: HookBase,
    pargs: argparse.Namespace,
    use_prompts: bool,
):
    cmd_defs = hook_obj.get_command_definitions()
    if pargs.cmd_path is None:
        if use_prompts:
            print("No command path (-p) argument specified, prompting from user")
            def_params.cmd_path = prompt_cmd_path(
                cmd_defs,
                history=hook_obj.get_cmd_history(),
                compl_style=params.app_params.compl_style,
            )
            print(f"Using command path: {def_params.cmd_path}")
    else:
        def_params.cmd_path = pargs.cmd_path


def args_to_params_generic(
    pargs: argparse.Namespace,
    params: SetupParams,
    hook_obj: HookBase,
    use_prompts: bool,
    assign_com_if: bool,
):
    if pargs.gui is None:
        params.app_params.use_gui = False
    else:
        params.app_params.use_gui = pargs.gui
    if pargs.com_if is None or pargs.com_if == CoreComInterfaces.UNSPECIFIED.value:
        assert hook_obj.cfg_path is not None
        params.com_if_id = determine_com_if(
            hook_obj.get_com_if_dict(), hook_obj.cfg_path, use_prompts
        )
    else:
        params.com_if_id = pargs.com_if
    if assign_com_if:
        params.com_if = hook_obj.assign_communication_interface(params.com_if_id)


def cfdp_args_to_cfdp_params(pargs: argparse.Namespace, cfdp_params: CfdpParams):
    """Convert the argument parser CFDP arguments provided by this library to the internalized
    :py:class:`tmtccmd.config.defs.CfdpParams` dataclass."""
    if hasattr(pargs, "source"):
        cfdp_params.source_file = pargs.source
    if hasattr(pargs, "target"):
        cfdp_params.dest_file = pargs.target
    if hasattr(pargs, "no_closure"):
        cfdp_params.closure_requested = not pargs.no_closure
    if hasattr(pargs, "type"):
        if pargs.type in ["0", "nak"]:
            cfdp_params.transmission_mode = TransmissionMode.UNACKNOWLEDGED
        elif pargs.type in ["1", "ack"]:
            cfdp_params.transmission_mode = TransmissionMode.ACKNOWLEDGED
    if hasattr(pargs, "proxy"):
        cfdp_params.proxy_op = pargs.proxy


def args_to_all_params_for_cfdp(
    pargs: argparse.Namespace,
    params: SetupParams,
    cfdp_params: CfdpParams,
    hook_obj: HookBase,
    use_prompts: bool,
    assign_com_if: bool,
):
    """Helper function to convert CFDP command line arguments to the setup parameters."""
    args_to_params_generic(
        pargs=pargs,
        params=params,
        hook_obj=hook_obj,
        use_prompts=use_prompts,
        assign_com_if=assign_com_if,
    )
    cfdp_args_to_cfdp_params(pargs, cfdp_params)
    # TODO: Listener mode is also relevant.
    #       Basically, if -l is specified, use listener after mqueue mode or right aways when
    #       no transaction parameter are specified
    #       Not sure if one queue mode is relevant here. A file transfer might always be split up
    #       in multiple queue fragments and might require feedback before finishing properly.
    params.backend_params.mode = CoreModeConverter.get_str(
        CoreModeList.MULTI_INTERACTIVE_QUEUE_MODE
    )
    if pargs.delay is None:
        params.cmd_params.delay = 0.4
    else:
        params.cmd_params.delay = float(pargs.delay)


def args_to_all_params_tmtc(
    pargs: argparse.Namespace,
    params: SetupParams,
    def_tmtc_params: TreeCommandingParams,
    hook_obj: HookBase,
    use_prompts: bool,
    assign_com_if: bool = True,
):
    """This function converts command line arguments to the internalized setup parameters.

    It is recommended to use the :py:class:`PostArgsParsingHelper` class to do this instead of
    calling this function directly.

    If some arguments are unspecified, they are set here with (variable) default values.

    :param pargs: Parsed arguments from calling parse method.
    :param params: Setup parameter object which will be set by this function.
    :param hook_obj:
    :param def_tmtc_params:
    :param use_prompts: Specify whether terminal prompts are allowed to retrieve unspecified
        arguments. For something like a GUI, it might make sense to disable this
    :param assign_com_if: Specifies whether this function should try to determine the COM interface
        from the specified key.
    :return: None
    """
    params.backend_params.listener = pargs.listener
    args_to_params_generic(
        pargs=pargs,
        params=params,
        hook_obj=hook_obj,
        use_prompts=use_prompts,
        assign_com_if=assign_com_if,
    )
    params.cmd_params.print_tree = False
    if pargs.print_tree is not None:
        params.cmd_params.print_tree = True
        for arg in pargs.print_tree:
            if "b" in arg:
                params.cmd_params.tree_print_with_description = False
            if arg.isdigit():
                params.cmd_params.tree_print_max_depth = int(arg)
    mode_set_explicitely = False
    if pargs.mode is None:
        params.mode = CoreModeConverter.get_str(CoreModeList.ONE_QUEUE_MODE)
    else:
        mode_set_explicitely = True
        params.mode = pargs.mode
    if (
        params.backend_params.listener
        and (not pargs.cmd_path)
        and not mode_set_explicitely
        and (not pargs.prompt_proc)
    ):
        params.mode = CoreModeConverter.get_str(CoreModeList.LISTENER_MODE)
    if pargs.delay is None:
        if params.backend_params.mode == CoreModeConverter.get_str(CoreModeList.ONE_QUEUE_MODE):
            params.cmd_params.delay = 4.0
        else:
            params.cmd_params.delay = 0.0
    else:
        params.cmd_params.delay = float(pargs.delay)
    if (
        params.mode != CoreModeConverter.get_str(CoreModeList.LISTENER_MODE)
        and not params.cmd_params.print_tree
    ):
        determine_cmd_path(
            params=params,
            hook_obj=hook_obj,
            use_prompts=use_prompts,
            pargs=pargs,
            def_params=def_tmtc_params,
        )


def perform_tree_printout(cmd_params: CommandingParams, cmd_def_tree: CmdTreeNode):
    if cmd_params.tree_print_with_description:
        info_str = "with full descriptions"
    else:
        info_str = "without descriptions"
    if cmd_params.tree_print_max_depth is not None:
        info_str += f" and maximum depth {cmd_params.tree_print_max_depth}"
    print(f"Printing command tree {info_str}:")
    print(
        cmd_def_tree.str_for_tree(
            cmd_params.tree_print_with_description,
            cmd_params.tree_print_max_depth,
        )
    )


class PreArgsParsingWrapper:
    """This class can be used to simplify parsing all tmtccmd CLI arguments.

    It wraps a parent parser and an argument parser but is also able to create default parsers.
    The :py:meth:`parse` method can be used to convert
    this parse the CLI arguments and then create a :py:class:`PostArgsParsingWrapper` to process
    these arguments.

    Please note that the parent parser and argument parser field have to be set or created
    first after creating this wrapper. They can be created by calling
    :py:meth:`create_default_parent_parser` and the :py:meth:`create_default_parser`, but the
    second function requires the parent parser to be set or created first.
    """

    def __init__(self):
        self.parent_parser: Optional[argparse.ArgumentParser] = None
        self.args_parser: Optional[argparse.ArgumentParser] = None
        self.print_known_args = False
        self.print_unknown_args = False
        self._parse_was_called = False
        self._monkey_patch_missing_subparser = False

    def create_default_parent_parser(self):
        """Create a default parent parser, which contains common flags for all possible
        tmtccmd submodules. For example, both the cfdp and default tmtc submodule could contain
        these common flags. The user can extend or modify the parent parser after it was created.
        """
        self.parent_parser = argparse.ArgumentParser(add_help=False)
        add_default_com_if_arguments(self.parent_parser)
        add_generic_arguments(self.parent_parser)

    def create_default_parser(self):
        """Create the default parser. Requires a valid parent parser containing common flags,
        The user can create or modify the parser after it was created. This function requires
        a valid parent parser to be set or created via :py:meth:`create_default_parent_parser`.
        """
        if self.parent_parser is None:
            raise ValueError(
                "Create parent parser with create_default_parent_parser or assign a "
                "parent_parser first"
            )
        self.args_parser = create_default_args_parser(self.parent_parser, descript_txt=None)

    def parse(self, hook_obj: HookBase, setup_params: SetupParams) -> PostArgsParsingWrapper:
        """Parses the set parser by calling the
        :py:meth:`argparse.ArgumentParser.parse_known_args` method internally and returns the
        :py:class:`PostArgsParsingWrapper` to simplify processing the arguments.
        """
        if self.parent_parser is None:
            raise ValueError("Create parser with create_default_parser or assign a parser first")
        """Parse all CLI arguments with the given argument parser"""
        patched_args = None
        if self._monkey_patch_missing_subparser:
            if (
                len(sys.argv) > 1
                and sys.argv[1] not in ["cfdp", "tmtc", "-h", "--help"]
                or len(sys.argv) == 1
            ):
                print("No procedure type specified, inserting 'tmtc' into passed" " arguments")
                patched_args = ["tmtc"]
                patched_args.extend(sys.argv[1:])
        if patched_args is None:
            patched_args = sys.argv[1:]
        assert self.args_parser is not None
        args_raw, unknown_args = parse_default_tmtccmd_input_arguments(
            args=patched_args,
            parser=self.args_parser,
            print_known_args=self.print_known_args,
            print_unknown_args=self.print_unknown_args,
        )
        return PostArgsParsingWrapper(
            args_raw=args_raw,
            unknown_args=unknown_args,
            hook_obj=hook_obj,
            params=setup_params,
        )

    def add_def_proc_args(self):
        """Add the default tmtc procedure parameters to the default parser. This includes
        the service and operation code flags."""
        self._check_arg_parser()
        assert self.args_parser is not None
        add_tree_commanding_arguments(self.args_parser)

    def add_cfdp_args(self):
        """Add the default CFDP procedure parameters to the default parser."""
        self._check_arg_parser()
        assert self.args_parser is not None
        add_cfdp_procedure_arguments(self.args_parser)

    def _check_arg_parser(self):
        if self.args_parser is None:
            raise ValueError(
                "Please build an argument parser first using the create_default_parser "
                "function or assigning args_parser"
            )

    def add_def_proc_and_cfdp_as_subparsers(
        self,
    ) -> Tuple[argparse.ArgumentParser, argparse.ArgumentParser]:
        """Add the default tmtc and cfdp procedure as subparsers."""
        self._monkey_patch_missing_subparser = True
        subparser = self.args_parser.add_subparsers(dest="proc_type")
        tmtc_parser = subparser.add_parser(
            "tmtc",
            help=(
                "Default TMTC Procedure Mode.\nDefault if no positional argument is" " specified"
            ),
            description=(
                "Default TMTC Procedure Mode using a Service and Operation "
                "Code Command Tuple to dispatch commands"
            ),
            formatter_class=argparse.RawTextHelpFormatter,
            parents=[self.parent_parser],
        )
        add_tree_commanding_arguments(tmtc_parser)
        cfdp_descrip = "CCSDS CFDP File Transfer"
        cfdp_parser = subparser.add_parser(
            "cfdp",
            help=cfdp_descrip,
            description=cfdp_descrip,
            formatter_class=argparse.RawTextHelpFormatter,
            parents=[self.parent_parser],
        )
        add_cfdp_procedure_arguments(cfdp_parser)
        return tmtc_parser, cfdp_parser


class PostArgsParsingWrapper:
    """This helper class helps with the internalization of the parse arguments into the format
    expected by tmtccmd.

    :var assign_com_if: If this is set to True (default), the wrapper will try to create
        a :py:class:`tmtccmd.com.ComInterface` on the conversion methods.
    """

    def __init__(
        self,
        args_raw: argparse.Namespace,
        unknown_args: List[str],
        params: SetupParams,
        hook_obj: HookBase,
    ):
        """It is recommended to use :py:meth:`PreArgsParsingWrapper.parse` to retrieve an instance
        instead of using the constructor directly.

        :param args_raw:
        :param unknown_args:
        :param params:
        :param hook_obj:
        """
        self.args_raw = args_raw
        self.params = params
        self.unknown_args = unknown_args
        self.hook_obj = hook_obj
        self.assign_com_if_on_conversion_if_applicable = True

    @property
    def use_gui(self):
        """This only yields valid values if :py:meth:`parse` was called once"""
        return self.args_raw.gui

    def request_type_from_args(self) -> TcProcedureType:
        if hasattr(self.args_raw, "proc_type"):
            if self.args_raw.proc_type == "tmtc":
                return TcProcedureType.TREE_COMMANDING
            elif self.args_raw.proc_type == "cfdp":
                return TcProcedureType.CFDP
            else:
                raise ValueError(
                    'Procedure type argument destination unknown, should be "tmtc" or' ' "cfdp"'
                )
        else:
            return TcProcedureType.TREE_COMMANDING

    def set_params_with_prompts(self, proc_base: ProcedureParamsWrapper):
        self._set_params(proc_base, True)

    def set_params_without_prompts(self, proc_wrapper: ProcedureParamsWrapper):
        self._set_params(proc_wrapper, False)

    def _set_params(self, proc_base: ProcedureParamsWrapper, with_prompts: bool):
        param_type = self.request_type_from_args()
        if param_type == TcProcedureType.TREE_COMMANDING:
            def_proc_params = TreeCommandingParams(None)
            if with_prompts:
                self.set_tmtc_params_with_prompts(def_proc_params)
            else:
                self.set_tmtc_params_without_prompts(def_proc_params)
            proc_base.set_params(def_proc_params)
        elif param_type == TcProcedureType.CFDP:
            cfdp_params = CfdpParams()
            if with_prompts:
                self.set_cfdp_params_with_prompts(cfdp_params)
            else:
                self.set_cfdp_params_without_prompts(cfdp_params)
            proc_base.set_params(cfdp_params)

    def set_cfdp_params_with_prompts(self, cfdp_params: CfdpParams):
        self._set_cfdp_params(cfdp_params, True)

    def set_cfdp_params_without_prompts(self, cfdp_params: CfdpParams):
        self._set_cfdp_params(cfdp_params, False)

    def set_tmtc_params_with_prompts(self, tmtc_params: TreeCommandingParams):
        self._set_tmtc_params(tmtc_params, True)

    def set_tmtc_params_without_prompts(self, tmtc_params: TreeCommandingParams):
        """Set up the parameter object from the parsed arguments. This call auto-determines whether
        prompts should be used depending on whether the GUI flag was passed or not.

        :raise Value Error: Parse function call missing
        """
        self._set_tmtc_params(tmtc_params, False)

    def _set_tmtc_params(
        self,
        def_tmtc_params: TreeCommandingParams,
        use_prompts: bool,
    ):
        try:
            args_to_all_params_tmtc(
                pargs=self.args_raw,
                params=self.params,
                hook_obj=self.hook_obj,
                use_prompts=use_prompts,
                def_tmtc_params=def_tmtc_params,
                assign_com_if=self.assign_com_if_on_conversion_if_applicable,
            )
        except KeyboardInterrupt:
            raise KeyboardInterrupt(
                "Keyboard interrupt while converting CLI args to application parameters"
            )

    def _set_cfdp_params(self, cfdp_params: CfdpParams, use_prompts: bool):
        try:
            args_to_all_params_for_cfdp(
                pargs=self.args_raw,
                cfdp_params=cfdp_params,
                params=self.params,
                hook_obj=self.hook_obj,
                use_prompts=use_prompts,
                assign_com_if=self.assign_com_if_on_conversion_if_applicable,
            )
        except KeyboardInterrupt:
            raise KeyboardInterrupt(
                "Keyboard interrupt while converting CLI args to application parameters"
            )



import struct
from unittest import TestCase

from spacepackets.ecss import Ptc, PfcUnsigned, PfcSigned, PfcReal
from tmtccmd.pus.s20_fsfw_param_defs import (
    create_scalar_boolean_parameter,
    parse_scalar_param,
    create_scalar_double_parameter,
    create_scalar_float_parameter,
    create_scalar_i8_parameter,
    create_scalar_i16_parameter,
    create_scalar_i32_parameter,
    create_scalar_u16_parameter,
    create_scalar_u32_parameter,
    create_scalar_u8_parameter,
    create_vector_float_parameter,
    create_vector_double_parameter,
    create_matrix_float_parameter,
    create_matrix_double_parameter,
)


class TestSrv20(TestCase):
    def setUp(self) -> None:
        self.obj_id = bytes([0x01, 0x02, 0x03, 0x04])
        self.boolean_param = create_scalar_boolean_parameter(
            object_id=self.obj_id, domain_id=1, unique_id=5, parameter=True
        )

    def test_param_id(self):
        expected_u32 = (1 << 24) | (5 << 16)
        param_id_u32 = self.boolean_param.param_id.as_u32()
        self.assertEqual(expected_u32, param_id_u32)

    def test_basic_boolean_param(self):
        self.assertEqual(self.boolean_param.ptc, Ptc.UNSIGNED)
        self.assertEqual(self.boolean_param.pfc, PfcUnsigned.ONE_BYTE)
        app_data_raw = self.boolean_param.pack()
        self.assertEqual(self.obj_id, app_data_raw[0:4])
        self.assertEqual(app_data_raw[4], 1)
        self.assertEqual(app_data_raw[5], 5)
        self.assertEqual(struct.unpack("!H", app_data_raw[6:8])[0], 0)
        self.assertEqual(app_data_raw[8], Ptc.UNSIGNED)
        self.assertEqual(app_data_raw[9], PfcUnsigned.ONE_BYTE)
        self.assertEqual(app_data_raw[10], 1)
        self.assertEqual(app_data_raw[11], 1)

    def test_float_param(self):
        double_param = create_scalar_float_parameter(
            self.obj_id, domain_id=12, unique_id=25, parameter=-12.0
        )
        raw = double_param.pack()
        self.assertEqual(len(raw), 16)

    def test_double_param(self):
        double_param = create_scalar_double_parameter(
            self.obj_id, domain_id=12, unique_id=25, parameter=3923929.0
        )
        raw = double_param.pack()
        self.assertEqual(len(raw), 20)
        double_val = struct.unpack("!d", raw[12:20])[0]
        self.assertTrue(abs(double_val - 3923929.0) < 0.00001)

    def test_param_extraction_bool(self):
        param = parse_scalar_param(self.boolean_param)
        self.assertEqual(param, 1)
        self.assertEqual(self.boolean_param.parse_scalar_param(), 1)

    def test_param_extraction_float(self):
        double_param = create_scalar_float_parameter(
            self.obj_id, domain_id=12, unique_id=25, parameter=-12.0
        )
        param = parse_scalar_param(double_param)
        self.assertTrue(abs(param + 12.0) < 0.00001)
        self.assertTrue(abs(double_param.parse_scalar_param() + 12.0) < 0.00001)

    def test_param_extraction_double(self):
        double_param = create_scalar_double_parameter(
            self.obj_id, domain_id=12, unique_id=25, parameter=3932.32
        )
        param = parse_scalar_param(double_param)
        self.assertTrue(abs(param - 3932.32) < 0.00001)
        self.assertTrue(abs(double_param.parse_scalar_param() - 3932.32) < 0.00001)

    def test_u8_param(self):
        u8_param = create_scalar_u8_parameter(
            self.obj_id, domain_id=25, unique_id=30, parameter=255
        )
        self.assertEqual(u8_param.param_raw[0], 255)
        self.assertEqual(u8_param.ptc, Ptc.UNSIGNED)
        self.assertEqual(u8_param.pfc, PfcSigned.ONE_BYTE)
        self.assertEqual(u8_param.rows, 1)
        self.assertEqual(u8_param.columns, 1)
        param = parse_scalar_param(u8_param)
        self.assertEqual(param, 255)

    def test_u16_param(self):
        u16_param = create_scalar_u16_parameter(
            self.obj_id, domain_id=25, unique_id=30, parameter=60000
        )
        self.assertEqual(u16_param.ptc, Ptc.UNSIGNED)
        self.assertEqual(u16_param.pfc, PfcUnsigned.TWO_BYTES)
        self.assertEqual(u16_param.rows, 1)
        self.assertEqual(u16_param.columns, 1)
        param = parse_scalar_param(u16_param)
        self.assertEqual(param, 60000)

    def test_u32_param(self):
        u32_param = create_scalar_u32_parameter(
            self.obj_id, domain_id=25, unique_id=30, parameter=79932932
        )
        self.assertEqual(u32_param.ptc, Ptc.UNSIGNED)
        self.assertEqual(u32_param.pfc, PfcUnsigned.FOUR_BYTES)
        self.assertEqual(u32_param.rows, 1)
        self.assertEqual(u32_param.columns, 1)
        param = parse_scalar_param(u32_param)
        self.assertEqual(param, 79932932)

    def test_i8_param(self):
        i8_param = create_scalar_i8_parameter(
            self.obj_id, domain_id=25, unique_id=30, parameter=-22
        )
        self.assertEqual(struct.unpack("!b", i8_param.param_raw)[0], -22)
        self.assertEqual(i8_param.ptc, Ptc.SIGNED)
        self.assertEqual(i8_param.pfc, PfcSigned.ONE_BYTE)
        self.assertEqual(i8_param.rows, 1)
        self.assertEqual(i8_param.columns, 1)
        param = parse_scalar_param(i8_param)
        self.assertEqual(param, -22)

    def test_i16_param(self):
        i16_param = create_scalar_i16_parameter(
            self.obj_id, domain_id=25, unique_id=30, parameter=-30000
        )
        self.assertEqual(struct.unpack("!h", i16_param.param_raw)[0], -30000)
        self.assertEqual(i16_param.ptc, Ptc.SIGNED)
        self.assertEqual(i16_param.pfc, PfcSigned.TWO_BYTES)
        self.assertEqual(i16_param.rows, 1)
        self.assertEqual(i16_param.columns, 1)
        param = parse_scalar_param(i16_param)
        self.assertEqual(param, -30000)

    def test_i32_param(self):
        i32_param = create_scalar_i32_parameter(
            self.obj_id, domain_id=25, unique_id=30, parameter=-70000
        )
        self.assertEqual(struct.unpack("!i", i32_param.param_raw)[0], -70000)
        self.assertEqual(i32_param.ptc, Ptc.SIGNED)
        self.assertEqual(i32_param.pfc, PfcSigned.FOUR_BYTES)
        self.assertEqual(i32_param.rows, 1)
        self.assertEqual(i32_param.columns, 1)
        param = parse_scalar_param(i32_param)
        self.assertEqual(param, -70000)

    def test_double_vec(self):
        double_vec = [3.0, -52.4, 38.32]
        double_vec_param = create_vector_double_parameter(
            self.obj_id, domain_id=30, unique_id=93, parameters=double_vec
        )
        self.assertEqual(double_vec_param.rows, 1)
        self.assertEqual(double_vec_param.columns, 3)
        self.assertEqual(double_vec_param.ptc, Ptc.REAL)
        self.assertEqual(double_vec_param.pfc, PfcReal.DOUBLE_PRECISION_IEEE)
        self.assertEqual(len(double_vec_param.param_raw), 24)
        self.assertTrue(abs(struct.unpack("!d", double_vec_param.param_raw[0:8])[0] - 3.0) < 0.0001)
        self.assertTrue(
            abs(struct.unpack("!d", double_vec_param.param_raw[8:16])[0]) - 52.4 < 0.0001
        )
        self.assertTrue(
            abs(struct.unpack("!d", double_vec_param.param_raw[16:24])[0] - 38.32) < 0.0001
        )

    def test_float_vec(self):
        float_vec = [3.0, -52.4, 38.32]
        double_vec_param = create_vector_float_parameter(
            self.obj_id, domain_id=30, unique_id=80, parameters=float_vec
        )
        self.assertEqual(double_vec_param.rows, 1)
        self.assertEqual(double_vec_param.columns, 3)
        self.assertEqual(double_vec_param.ptc, Ptc.REAL)
        self.assertEqual(double_vec_param.pfc, PfcReal.FLOAT_SIMPLE_PRECISION_IEEE)
        self.assertEqual(len(double_vec_param.param_raw), 12)
        self.assertTrue(abs(struct.unpack("!f", double_vec_param.param_raw[0:4])[0] - 3.0) < 0.0001)
        self.assertTrue(
            abs(struct.unpack("!f", double_vec_param.param_raw[4:8])[0]) - 52.4 < 0.0001
        )
        self.assertTrue(
            abs(struct.unpack("!f", double_vec_param.param_raw[8:12])[0]) - 38.32 < 0.0001
        )

    def test_float_matrix(self):
        float_matrix = [
            [3.0, -52.4, 28.32],
            [94924.42, -30232.32, 9393.2],
        ]
        float_matrix_param = create_matrix_float_parameter(
            self.obj_id, domain_id=30, unique_id=80, parameters=float_matrix
        )
        self.assertEqual(float_matrix_param.rows, 2)
        self.assertEqual(float_matrix_param.columns, 3)
        self.assertEqual(float_matrix_param.ptc, Ptc.REAL)
        self.assertEqual(float_matrix_param.pfc, PfcReal.FLOAT_SIMPLE_PRECISION_IEEE)
        self.assertEqual(len(float_matrix_param.param_raw), 24)
        self.assertTrue(
            abs(struct.unpack("!f", float_matrix_param.param_raw[0:4])[0] - 3.0) < 0.0001
        )
        self.assertTrue(
            abs(struct.unpack("!f", float_matrix_param.param_raw[4:8])[0]) - 52.4 < 0.0001
        )
        self.assertTrue(
            abs(struct.unpack("!f", float_matrix_param.param_raw[8:12])[0]) - 28.32 < 0.0001
        )
        self.assertTrue(
            abs(struct.unpack("!f", float_matrix_param.param_raw[12:16])[0]) - 94924.42 < 0.01
        )
        self.assertTrue(
            abs(struct.unpack("!f", float_matrix_param.param_raw[16:20])[0]) - 30232.32 < 0.01
        )
        self.assertTrue(
            abs(struct.unpack("!f", float_matrix_param.param_raw[20:24])[0]) - 9393.2 < 0.001
        )

    def test_double_matrix(self):
        double_matrix = [
            [3.0, -52.4, 28.32],
            [94924.42, -30232.32, 9393.2],
        ]
        float_matrix_param = create_matrix_double_parameter(
            self.obj_id, domain_id=30, unique_id=80, parameters=double_matrix
        )
        self.assertEqual(float_matrix_param.rows, 2)
        self.assertEqual(float_matrix_param.columns, 3)
        self.assertEqual(float_matrix_param.ptc, Ptc.REAL)
        self.assertEqual(float_matrix_param.pfc, PfcReal.DOUBLE_PRECISION_IEEE)
        self.assertEqual(len(float_matrix_param.param_raw), 48)
        self.assertTrue(
            abs(struct.unpack("!d", float_matrix_param.param_raw[0:8])[0] - 3.0) < 0.0001
        )
        self.assertTrue(
            abs(struct.unpack("!d", float_matrix_param.param_raw[8:16])[0]) - 52.4 < 0.0001
        )
        self.assertTrue(
            abs(struct.unpack("!d", float_matrix_param.param_raw[16:24])[0]) - 28.32 < 0.0001
        )
        self.assertTrue(
            abs(struct.unpack("!d", float_matrix_param.param_raw[24:32])[0]) - 94924.42 < 0.0001
        )
        self.assertTrue(
            abs(struct.unpack("!d", float_matrix_param.param_raw[32:40])[0]) - 30232.32 < 0.0001
        )
        self.assertTrue(
            abs(struct.unpack("!d", float_matrix_param.param_raw[40:48])[0]) - 9393.2 < 0.00001
        )



# -*- coding: utf-8 -*-
"""Communication module. Provides generic abstraction for communication and commonly used
concrete implementations."""

from abc import abstractmethod, ABC
from typing import Any, List, Optional

from tmtccmd.tmtc.common import TelemetryListT


class ReceptionDecodeError(Exception):
    """Generic decode error which can also wrap the exception thrown by other libraries."""

    def __init__(self, msg: str, custom_exception: Optional[Exception]):
        super().__init__(msg)
        self.custom_exception = custom_exception


class SendError(Exception):
    """Generic send error which can also wrap the exception thrown by other libraries."""

    def __init__(self, msg: str, custom_exception: Optional[Exception]):
        super().__init__(msg)
        self.custom_exception = custom_exception


class ComInterface(ABC):
    """Generic form of a communication interface to separate communication logic from
    the underlying interface.
    """

    @property
    @abstractmethod
    def id(self) -> str:
        pass

    @abstractmethod
    def initialize(self, args: Any = 0) -> Any:
        """Perform initializations step which can not be done in constructor or which require
        returnvalues.
        """

    @abstractmethod
    def open(self, args: Any = 0):
        """Opens the communication interface to allow communication.

        :return:
        """

    @abstractmethod
    def is_open(self) -> bool:
        """Can be used to check whether the communication interface is open. This is useful if
        opening a COM interface takes a longer time and is non-blocking
        """

    @abstractmethod
    def close(self, args: Any = 0):
        """Closes the ComIF and releases any held resources (for example a Communication Port).

        :return:
        """

    @abstractmethod
    def send(self, data: bytes):
        """Send raw data.

        :raises SendError: Sending failed for some reason.
        """

    @abstractmethod
    def receive(self, parameters: Any = 0) -> List[bytes]:
        """Returns a list of received packets. The child class can use a separate thread to poll for
        the packets or use some other mechanism and container like a deque to store packets
        to be returned here.

        :param parameters:
        :raises ReceptionDecodeError: If the underlying COM interface uses encoding and
            decoding and the decoding fails, this exception will be returned.
        :return:
        """
        packet_list = []
        return packet_list

    @abstractmethod
    def data_available(self, timeout: float, parameters: Any = 0) -> int:
        """Check whether TM packets are available.

        :param timeout: Can be used to block on available data if supported by the specific
            communication interface.
        :param parameters: Can be an arbitrary parameter.
        :raises ReceptionDecodeError: If the underlying COM interface uses encoding and
            decoding when determining the number of available packets, this exception can be
            thrown on decoding errors.
        :return: 0 if no data is available, number of packets otherwise.
        """



from unittest import TestCase

from spacepackets.ecss import PfcUnsigned, Ptc, PusTc, PusService
from tmtccmd.pus.s20_fsfw_param_defs import Parameter, CustomSubservice
from tmtccmd.pus.s20_fsfw_param import (
    create_load_param_cmd,
    create_scalar_boolean_parameter,
)


class TestSrv20FsfwTc(TestCase):
    def setUp(self):
        self.obj_id = bytes([0x01, 0x02, 0x03, 0x04])
        self.boolean_param = create_scalar_boolean_parameter(
            object_id=self.obj_id, domain_id=1, unique_id=5, parameter=True
        )
        self.assertEqual(self.boolean_param.ptc, Ptc.UNSIGNED)
        self.assertEqual(self.boolean_param.pfc, PfcUnsigned.ONE_BYTE)
        self.tc = create_load_param_cmd(self.boolean_param, 0x05)

    def test_basic(self):
        # 12 bytes of generic parameter header + 1 byte parameter itself
        self.assertEqual(len(self.tc.app_data), 13)
        self.assertEqual(self.tc.apid, 0x05)
        self.assertEqual(self.tc.service, PusService.S20_PARAMETER)
        self.assertEqual(self.tc.subservice, CustomSubservice.TC_LOAD)

    def test_unpack(self):
        raw_tc = self.tc.pack()
        unpacked = PusTc.unpack(raw_tc)
        boolean_param_conv_back = Parameter.unpack(unpacked.app_data)
        self.assertEqual(boolean_param_conv_back, self.boolean_param)



"""`FSFW <https://egit.irs.uni-stuttgart.de/fsfw/fsfw>`_ specific PUS event support"""

from .s5_fsfw_event_defs import *  # noqa re-export
from .tm.s5_fsfw_event import *  # noqa re-export
from .tc.s5_event import *  # noqa re-export



from __future__ import annotations

import enum
import copy
import os
from typing import Any, Dict, List, Optional, Set, Tuple, Union

from deprecated.sphinx import deprecated


class TreePart(enum.Enum):
    EDGE = "├──"
    LINE = "│  "
    CORNER = "└──"
    BLANK = "   "


class DepthInfo:
    def __init__(
        self,
        depth: int,
        last_child: bool,
        max_depth: Optional[int] = None,
        layer_is_last_set: Optional[Set[int]] = None,
    ) -> None:
        self.depth = depth
        self.last_child = last_child
        if layer_is_last_set is None:
            self.set_of_layers_where_child_is_last = set()
        else:
            self.set_of_layers_where_child_is_last = layer_is_last_set
        self.max_depth = max_depth

    def is_layer_for_last_child(self, depth: int) -> bool:
        return depth in self.set_of_layers_where_child_is_last

    def set_layer_is_last_child(self, depth: int):
        self.set_of_layers_where_child_is_last.add(depth)

    def clear_layer_is_last_child(self, depth: int):
        self.set_of_layers_where_child_is_last.remove(depth)


class CmdTreeNode:
    """The command tree node is the primary data structure used to specify the command
    structure in a way it can be used by framework components.

    The node class provides an API which allows to build a tree of command nodes. Generally, a full
    path from the root node to a leaf will be the command identifier or command path for executing
    a certain command or procedure.

    You can create the root node using the :py:meth:`CmdTreeNode.root_node` class method. After
    that children can be appended to the nodes using the :py:meth:`CmdTreeNode.add_child` method.

    You can use square bracket operator to access the children of a node directly. For example,
    if a node with the name ``test_node`` has the child ``event``, you could use
    ``test_node["event"]`` to access the child node.

    """

    def __init__(
        self,
        name: str,
        description: str,
        parent: Optional[CmdTreeNode] = None,
        hide_children_for_print: bool = False,
        hide_children_which_are_leaves: bool = False,
    ) -> None:
        """
        Parameters
        ------------

        name
            Name of the node, which will be also part of the command path when picking a path
            through the tree.
        description
            Additional description for the node.
        parent
            Parent of the node. Generally, this does not need to be set, as it will be set
            correctly when using the :py:meth:`CmdTreeNode.add_child` method.
        hide_children_for_print
            For large tree, it can make sense to hide the children for a regular printout of the
            tree. This field allows to do this.
        hide_children_which_are_leaves
            For large tree, it can make sense to hide the children which are leaves for a regular
            printout of the tree. This field allows to do this. This field is overriden by the
            strong ``hide_children_for_print`` field.

        """
        self.name = name
        self.description = description
        self.parent: Optional[CmdTreeNode] = parent
        self.children: Dict[str, CmdTreeNode] = {}
        self.hide_children_for_print = hide_children_for_print
        self.hide_children_which_are_leaves = hide_children_which_are_leaves

    @classmethod
    def root_node(cls) -> CmdTreeNode:
        return cls(name="/", description="Root Node", parent=None)

    def __getitem__(self, arg):
        return self.children[arg]

    def add_child(self, child: CmdTreeNode):
        """Add a child to the node. This will also assign the parent class of the child to
        the current node."""
        child.parent = self
        self.children.update({child.name: child})

    def is_leaf(self) -> bool:
        """A leaf is a node which has no children."""
        return len(self.children) == 0

    def contains_path(self, path: str) -> bool:
        """Check whether a full slash separated command path is contained within
        the command tree."""
        if path == "":
            return False
        return self.contains_path_from_node_list(path.split("/"))

    def contains_path_from_node_list(self, node_name_list: List[str]) -> bool:
        """Check whether the given list of nodes are contained within the command tree."""
        if len(node_name_list) == 0:
            return False
        # Only root node.
        if len(node_name_list) == 2 and node_name_list == ["", ""]:
            return True
        if node_name_list[0] == "":
            # Cut off empty string left of root node /
            node_name_list = node_name_list[1:]
        for child in self.children.values():
            if node_name_list[0] == child.name:
                # This is the last node.
                if len(node_name_list) == 1:
                    return True
                return child.contains_path_from_node_list(node_name_list[1:])
        return False

    def extract_subnode(self, path: str) -> Optional[CmdTreeNode]:
        """Extract a subnode given a relative path."""
        if path == "":
            return None
        return self.extract_subnode_by_node_list(path.split("/"))

    def extract_subnode_by_node_list(self, node_list: List[str]) -> Optional[CmdTreeNode]:
        """Extract a subnode given a list which would form a relative path if it were joined
        using slashes."""
        if not self.contains_path_from_node_list(node_list):
            return None
        if len(node_list) == 1 and self.children[node_list[0]] is not None:
            return self.children[node_list[0]]
        return self.children[node_list[0]].extract_subnode_by_node_list(node_list[1:])

    @property
    def name_dict(self) -> Dict[str, Optional[Dict[str, Any]]]:
        """Returns a nested dictionary where the key is always the name of the node, and the
        value is one nested name dictionary for each child node."""
        children_dict = {}
        if self.children:
            for child in self.children.values():
                children_dict.update(child.name_dict)
            return {self.name: children_dict}
        return {self.name: None}

    def str_for_tree(
        self,
        with_description: bool,
        max_depth: Optional[int] = None,
        show_hidden_elements: bool = False,
    ) -> str:
        """Retrieve the a human readable printout of the tree.

        Parameters
        ------------

        with_description
            Display descriptions right to the tree.
        max_depth
            Entries will be cut-off at the specified depth. None can be specified to print all
            depths.
        show_hidden_elements
            Overrides the hide argument of command tree nodes.
        """
        return self.__str_for_depth(
            with_description,
            DepthInfo(depth=0, last_child=False, max_depth=max_depth),
            show_hidden_elements=show_hidden_elements,
        )

    def __str_for_depth(
        self,
        with_description: bool,
        depth_info: DepthInfo,
        show_hidden_elements: bool = False,
    ) -> str:
        # If we are at a larger depth than 0, we want to prepend the name using a special
        # format. Example:
        #
        # Example 1:
        # /
        # ├── ACS
        # │  └── ACS_CTRL
        # └── TCS
        #
        # Example 2:
        # /
        # ├── ACS
        # │  └── ACS_CTRL
        # └── TCS
        #    └── TCS_CTRL
        #
        # Example 3:
        # /
        # ├── ACS
        # │  ├── ACS_CTRL
        # │  └── MGM_0
        # │     └── UPDATE_CFG
        # ├── TCS
        # │  ├── TCS_CTRL
        # │  └── PT1000_0
        # └── PING
        line = CmdTreeNode._create_leading_tree_part(depth_info)
        if depth_info.max_depth is not None and depth_info.depth > depth_info.max_depth:
            line += f"... (cut-off, maximum depth {depth_info.max_depth}){os.linesep}"
            return line
        else:
            line += self.name
            if with_description:
                line = f"{line.ljust(35)} [ " + self.description + " ]"
        string = line + os.linesep
        return self._handle_children_printout(
            string, with_description, depth_info, show_hidden_elements
        )

    @staticmethod
    def _create_leading_tree_part(depth_info: DepthInfo) -> str:
        line = ""
        for i in range(depth_info.depth):
            if i == depth_info.depth - 1:
                if depth_info.last_child:
                    line += TreePart.CORNER.value + " "
                else:
                    line += TreePart.EDGE.value + " "
            else:
                if depth_info.is_layer_for_last_child(i):
                    line += TreePart.BLANK.value
                else:
                    line += TreePart.LINE.value
        return line

    def _handle_children_printout(
        self,
        string: str,
        with_description: bool,
        depth_info: DepthInfo,
        show_hidden_elements: bool,
    ) -> str:
        child_depth_info = DepthInfo(
            depth=depth_info.depth + 1,
            last_child=False,
            max_depth=depth_info.max_depth,
            layer_is_last_set=copy.copy(depth_info.set_of_layers_where_child_is_last),
        )
        if not show_hidden_elements and self.hide_children_for_print and len(self.children) > 0:
            CmdTreeNode._this_is_the_last_child(child_depth_info, depth_info)
            string += CmdTreeNode._create_supressed_children_line(
                child_depth_info, "children are hidden"
            )
        else:
            if not show_hidden_elements and self.hide_children_which_are_leaves:
                str_if_there_are_leaves = self._handle_children_output_suppressed_leaves(
                    with_description, depth_info, child_depth_info
                )
                if str_if_there_are_leaves is not None:
                    return string + str_if_there_are_leaves
            for idx, child in enumerate(self.children.values()):
                if idx == len(self.children) - 1:
                    CmdTreeNode._this_is_the_last_child(child_depth_info, depth_info)
                # Use recursion here to get the string for the subtree.
                string += child.__str_for_depth(
                    with_description=with_description,
                    depth_info=child_depth_info,
                    show_hidden_elements=show_hidden_elements,
                )
        return string

    def _handle_children_output_suppressed_leaves(
        self,
        with_description: bool,
        depth_info: DepthInfo,
        child_depth_info: DepthInfo,
    ) -> Optional[str]:
        string = ""
        children_which_are_not_leaves = []
        some_children_which_are_leaves = False
        # The children need to be sorted first: Children which are not leaves come first.
        for child in self.children.values():
            if child.is_leaf():
                some_children_which_are_leaves = True
                continue
            children_which_are_not_leaves.append(child.name)
        if not some_children_which_are_leaves:
            return None
        for child in children_which_are_not_leaves:
            string += self.children[child].__str_for_depth(with_description, child_depth_info)
        CmdTreeNode._this_is_the_last_child(child_depth_info, depth_info)
        string += CmdTreeNode._create_supressed_children_line(child_depth_info, "leaves are hidden")
        return string

    @staticmethod
    def _this_is_the_last_child(child_depth_info: DepthInfo, depth_info: DepthInfo):
        child_depth_info.last_child = True
        child_depth_info.set_layer_is_last_child(depth_info.depth)

    @staticmethod
    def _create_supressed_children_line(child_depth_info: DepthInfo, reason: str) -> str:
        line = CmdTreeNode._create_leading_tree_part(child_depth_info)
        line += f"... (cut-off, {reason}){os.linesep}"
        return line

    def __str__(self) -> str:
        return self.str_for_tree(False)


ServiceNameT = str
ServiceInfoT = str
OpCodeNameT = Union[str, List[str]]
OpCodeInfoT = str


@deprecated(
    version="8.0.0",
    reason="use the new command definition tree instead",
)
class OpCodeOptionBase:
    def __init__(self):
        pass


OpCodeDict = Dict[str, Tuple[OpCodeInfoT, OpCodeOptionBase]]


@deprecated(
    version="8.0.0",
    reason="use the new command definition tree instead",
)
class OpCodeEntry:
    def __init__(self):
        self._op_code_dict_num_keys: OpCodeDict = dict()
        self._op_code_dict_str_keys: OpCodeDict = dict()

    def add(
        self,
        keys: OpCodeNameT,
        info: str,
        options: OpCodeOptionBase = OpCodeOptionBase(),
    ):
        if isinstance(keys, str):
            keys = [keys]
        for key in keys:
            if key.isdigit():
                self._op_code_dict_num_keys.update({key: (info, options)})
            else:
                self._op_code_dict_str_keys.update({key: (info, options)})

    def sort_num_key_dict(self):
        self._op_code_dict_num_keys = {int(k): v for k, v in self._op_code_dict_num_keys.items()}

    def sort_text_key_dict(self):
        self._op_code_dict_str_keys = {k: v for k, v in sorted(self._op_code_dict_str_keys.items())}

    def info(self, op_code: str) -> Optional[str]:
        if op_code.isdigit():
            entry_tuple = self._op_code_dict_num_keys.get(op_code)
            if entry_tuple is not None:
                return entry_tuple[0]

        entry_tuple = self._op_code_dict_str_keys.get(op_code)
        if entry_tuple is not None:
            return entry_tuple[0]
        return None

    def __str__(self):
        return (
            f"Op codes with numeric keys: {self._op_code_dict_num_keys!r}), "
            f"op codes with text keys: {self._op_code_dict_str_keys!r}"
        )

    @property
    def op_code_dict_num_keys(self):
        return self._op_code_dict_num_keys

    @property
    def op_code_dict_str_keys(self):
        return self._op_code_dict_str_keys


# It is possible to specify a service without any op codes
ServiceDictValueT = Optional[Tuple[ServiceInfoT, OpCodeEntry]]
ServiceOpCodeDictT = Dict[ServiceNameT, ServiceDictValueT]


@deprecated(
    version="8.0.0",
    reason="use the new command definition tree instead",
)
class TmtcDefinitionWrapper:
    def __init__(self, init_defs: Optional[ServiceOpCodeDictT] = None):
        if init_defs is None:
            self.defs: ServiceOpCodeDictT = dict()
        else:
            self.defs = init_defs

    def __repr__(self):
        return f"{self.__class__.__name__}(init_defs={self.defs!r}"

    def add_service(
        self,
        name: str,
        info: str,
        op_code_entry: OpCodeEntry,
    ):
        self.defs.update({name: (info, op_code_entry)})

    def op_code_entry(self, service_name: str) -> Optional[OpCodeEntry]:
        srv_entry = self.defs.get(service_name)
        if srv_entry is not None:
            return srv_entry[1]
        return None

    def sort(self):
        self.defs = {key: self.defs[key] for key in sorted(self.defs.keys())}


REGISTER_CBS = set()


@deprecated(
    version="8.0.0",
    reason="use the new command definition tree instead",
)
def tmtc_definitions_provider(adder_func):
    """Function decorator which registers the decorated function to be a TMTC definition provider.
    The :py:func:`execute_tmtc_def_functions` function can be used to call all functions.

    It is expected that the decorated function takes the py:class:`TmTcDefWrapper` as the first
    argument. The user can pass any additional arguments as positional and keyword
    arguments.
    """
    global REGISTER_CBS

    def call_explicitely(defs: TmtcDefinitionWrapper, *args, **kwargs):
        if REGISTER_CBS and adder_func in REGISTER_CBS:
            REGISTER_CBS.remove(adder_func)
        print(f"{adder_func.__name__} called explicitely, not necessary")
        return adder_func(defs, args, kwargs)

    REGISTER_CBS.add(adder_func)
    return call_explicitely


@deprecated(
    version="8.0.0",
    reason="use the new command definition tree instead",
)
def call_all_definitions_providers(defs: TmtcDefinitionWrapper, *args, **kwargs):
    global REGISTER_CBS
    if REGISTER_CBS:
        for register_cb in REGISTER_CBS.copy():
            if args and kwargs:
                register_cb(defs, args, kwargs)
            if args and not kwargs:
                register_cb(defs, args)
            elif not args and not kwargs:
                register_cb(defs)
            REGISTER_CBS.remove(register_cb)



import enum


class FsfwHealth(enum.IntEnum):
    FAULTY = 0
    HEALTHY = 1
    EXTERNAL_CTRL = 2
    NEEDS_RECOVERY = 3
    PERMANENT_FAULTY = 4


def pack_set_health_cmd_data(object_id: bytes, health: FsfwHealth) -> bytearray:
    cmd = bytearray()
    cmd += object_id
    cmd.append(health)
    return cmd



from unittest import TestCase

from tmtccmd.util.obj_id import (
    ComponentIdU16,
    ComponentIdU32,
    ComponentIdU8,
)


class TestObjectId(TestCase):
    def test_basic(self):
        obj_id0 = ComponentIdU32(1, "Some Name")
        self.assertEqual(str(obj_id0), "Object ID Some Name with ID 0x00000001")
        self.assertEqual(obj_id0.__repr__(), "ComponentIdU32(object_id=1, name='Some Name')")
        self.assertEqual(obj_id0.as_bytes, bytes([0x00, 0x00, 0x00, 0x01]))
        self.assertEqual(obj_id0.as_hex_string, "0x00000001")
        self.assertEqual(int(obj_id0), 1)
        obj_id1 = ComponentIdU32(1, "Other Name")
        self.assertEqual(obj_id0, obj_id1)
        obj_from_raw = ComponentIdU32.from_bytes(obj_id0.as_bytes)
        self.assertEqual(obj_from_raw, obj_id0)
        with self.assertRaises(ValueError):
            ComponentIdU32.from_bytes(bytes())
        with self.assertRaises(ValueError):
            ComponentIdU32.from_bytes(bytes([0, 1, 2]))
        with self.assertRaises(ValueError):
            obj_id1.obj_id = -1

    def test_diff_types(self):
        obj_id_u8 = ComponentIdU8(1, "U8 ID 0")
        self.assertEqual(obj_id_u8.as_bytes, bytes([1]))
        self.assertEqual(obj_id_u8.as_hex_string, "0x01")
        self.assertEqual(obj_id_u8.byte_len, 1)
        obj_id_u16 = ComponentIdU16(2, "U16 ID 2")
        self.assertEqual(obj_id_u16.as_bytes, bytes([0, 2]))
        self.assertEqual(obj_id_u16.as_hex_string, "0x0002")
        self.assertEqual(obj_id_u16.byte_len, 2)
        obj_id_u32 = ComponentIdU32(1, "U32 ID 1")
        test_dict = dict()
        test_dict.update({obj_id_u8: obj_id_u8.name})
        test_dict.update({obj_id_u16: obj_id_u16.name})
        test_dict.update({obj_id_u32: obj_id_u32.name})
        self.assertEqual(len(test_dict), 3)
        obj_id_u8_from_raw = ComponentIdU8.from_bytes(obj_id_u8.as_bytes)
        self.assertEqual(obj_id_u8_from_raw, obj_id_u8)
        obj_id_u16_from_raw = ComponentIdU16.from_bytes(obj_id_u16.as_bytes)
        self.assertEqual(obj_id_u16_from_raw, obj_id_u16)
        obj_id_u32_from_raw = ComponentIdU32.from_bytes(obj_id_u32.as_bytes)
        self.assertEqual(obj_id_u32_from_raw, obj_id_u32)



from .obj_id import (
    ComponentIdU32,
    ComponentIdU16,
    ComponentIdU8,
    ComponentIdBase,
    ComponentIdMapping,
    ObjectIdU32,
    ObjectIdU16,
    ObjectIdU8,
    ObjectIdBase,
    ObjectIdMapping,
    ObjectIdDictT,
)
from .retval import RetvalDictT



import collections
import dataclasses
import enum
from enum import auto
import logging
import time
from typing import Optional

import serial


class SerialConfigIds(enum.Enum):
    SERIAL_PORT = auto()
    SERIAL_BAUD_RATE = auto()
    SERIAL_TIMEOUT = auto()
    SERIAL_COMM_TYPE = auto()
    SERIAL_FRAME_SIZE = auto()
    SERIAL_DLE_QUEUE_LEN = auto()
    SERIAL_DLE_MAX_FRAME_SIZE = auto()


class SerialCommunicationType(enum.Enum):
    """
    Right now, two serial communication methods are supported. One uses frames with a fixed size
    containing PUS packets and the other uses a simple ASCII based transport layer called DLE.
    If DLE is used, it is expected that the sender side encoded the packets with the DLE
    protocol. Any packets sent will also be encoded.
    """

    COBS = 0
    DLE_ENCODING = 2


@dataclasses.dataclass
class SerialCfg:
    com_if_id: str
    serial_port: str
    baud_rate: int
    serial_timeout: float


class SerialComBase:
    def __init__(
        self,
        logger: logging.Logger,
        ser_cfg: SerialCfg,
        ser_com_type: SerialCommunicationType,
    ):
        self.logger = logger
        self.ser_cfg = ser_cfg
        self.ser_com_type = ser_com_type
        self.serial: Optional[serial.Serial] = None

    def open_port(self):
        try:
            self.serial = serial.Serial(
                port=self.ser_cfg.serial_port,
                baudrate=self.ser_cfg.baud_rate,
                timeout=self.ser_cfg.serial_timeout,
            )
        except serial.SerialException:
            self.logger.error("Serial Port opening failure!")
            raise IOError

    def close_port(self):
        try:
            self.serial.close()
            self.serial = None
        except serial.SerialException:
            logging.warning("SERIAL Port could not be closed!")

    def is_port_open(self) -> bool:
        return self.serial is not None

    @staticmethod
    def data_available_from_queue(timeout: float, reception_buffer: collections.deque):
        elapsed_time = 0
        start_time = time.time()
        sleep_time = timeout / 3.0
        if timeout > 0:
            while elapsed_time < timeout:
                if reception_buffer:
                    return reception_buffer.__len__()
                elapsed_time = time.time() - start_time
                time.sleep(sleep_time)
        if reception_buffer:
            return reception_buffer.__len__()
        return 0



import socket
import threading
import time
from collections import deque
from typing import Optional
from unittest import TestCase

from spacepackets import PacketType
from spacepackets.ccsds import PacketId
from spacepackets.ecss import PusTelecommand, PusTelemetry
from tmtccmd.com.tcp import TcpSpacepacketsClient
from tmtccmd.com.tcpip_utils import EthAddr


LOCALHOST = "127.0.0.1"


class TestTcpIf(TestCase):
    def setUp(self) -> None:
        self.tcp_server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.tcp_server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # Let the OS assign a port.
        self.addr = (LOCALHOST, 0)
        self.tcp_server.bind(self.addr)
        # Update the address for the client.
        self.addr = self.addr[0], self.tcp_server.getsockname()[1]
        self.tcp_server.listen()
        self.expected_packet_id = PacketId(apid=0x22, sec_header_flag=True, ptype=PacketType.TM)
        self.base_data = bytes([0, 1, 2, 3])
        self.ping_cmd = PusTelecommand(service=17, subservice=1, apid=0x22)
        self.ping_reply = PusTelemetry(service=17, subservice=2, apid=0x22, timestamp=bytes())
        self.tcp_client = TcpSpacepacketsClient(
            "tcp",
            space_packet_ids=[self.expected_packet_id],
            target_address=EthAddr.from_tuple(self.addr),
            inner_thread_delay=0.05,
        )
        self.conn_socket: Optional[socket.socket] = None
        self.server_received_packets = deque()
        self.tcp_client.initialize()

    def test_monolithic(self):
        self._open()
        tcp_server = threading.Thread(target=self.tcp_server_thread, daemon=True)
        tcp_server.start()
        self._test_basic()
        self._test_send()
        self._test_recv()
        self._test_close_client()

    def tcp_server_thread(self):
        (conn_sock, addr_info) = self.tcp_server.accept()
        while True:
            (data_recv, sender_addr) = conn_sock.recvfrom(4096)
            if len(data_recv) == 0:
                try:
                    conn_sock.shutdown(socket.SHUT_RDWR)
                except OSError:
                    # Endpoint might already be closed, or no connection to it.
                    pass
                conn_sock.close()
                break
            else:
                if data_recv == self.ping_cmd.pack():
                    self.server_received_packets.appendleft(data_recv)
                    conn_sock.sendall(self.ping_reply.pack())
                else:
                    self.server_received_packets.appendleft(data_recv)

    def _test_basic(self):
        self.assertEqual(self.tcp_client.id, "tcp")

    def _test_send(self):
        self.tcp_client.send(self.base_data)
        time.sleep(0.2)
        self.assertEqual(len(self.server_received_packets), 1)
        packet = self.server_received_packets.pop()
        self.assertEqual(packet, self.base_data)

    def _test_close_client(self):
        self.tcp_client.close()

    def _test_recv(self):
        self.tcp_client.send(self.ping_cmd.pack())
        time.sleep(0.2)
        # Assert it arrived at the server
        self.assertEqual(len(self.server_received_packets), 1)
        packet = self.server_received_packets.pop()
        self.assertEqual(packet, self.ping_cmd.pack())
        # Now assert that a ping reply was sent back to the client if a ping command was sent
        self.assertEqual(self.tcp_client.data_available(), 1)
        recvd_packets = self.tcp_client.receive()
        self.assertEqual(len(recvd_packets), 1)
        self.assertEqual(recvd_packets[0], self.ping_reply.pack())

    def _open(self):
        self.tcp_client.open()
        self.assertTrue(self.tcp_client.is_open())

    def tearDown(self) -> None:
        self.tcp_server.close()
        self.tcp_client.close()



import os
from typing import Optional
from unittest import TestCase
from tmtccmd.config.tmtc import CmdTreeNode


class TestCmdDefTree(TestCase):
    def setUp(self) -> None:
        self.cmd_tree = CmdTreeNode.root_node()

    def base_tree(self):
        self.cmd_tree.add_child(CmdTreeNode("acs", "ACS Subsystem"))
        self.cmd_tree.add_child(CmdTreeNode("tcs", "TCS Subsystem"))

    def tree_with_two_layers(self):
        self.base_tree()
        self.cmd_tree.add_child(CmdTreeNode("ping", "Ping Command"))
        self.cmd_tree.children["acs"].add_child(CmdTreeNode("acs_ctrl", "ACS Controller"))

    def test_state(self):
        self.base_tree()
        self.cmd_tree.add_child(CmdTreeNode("ping", "Ping Command"))
        self.assertEqual(self.cmd_tree.name, "/")
        self.assertEqual(self.cmd_tree.children["acs"].name, "acs")
        self.assertEqual(self.cmd_tree.children["acs"].description, "ACS Subsystem")
        self.assertEqual(self.cmd_tree.children["acs"].children, {})
        assert self.cmd_tree.children["acs"].parent is not None
        self.assertEqual(self.cmd_tree.children["acs"].parent.name, "/")

        self.assertEqual(self.cmd_tree.children["tcs"].name, "tcs")
        self.assertEqual(self.cmd_tree.children["tcs"].description, "TCS Subsystem")
        self.assertEqual(self.cmd_tree.children["tcs"].children, {})
        assert self.cmd_tree.children["tcs"].parent is not None
        self.assertEqual(self.cmd_tree.children["tcs"].parent.name, "/")

        self.assertEqual(self.cmd_tree.children["ping"].name, "ping")
        self.assertEqual(self.cmd_tree.children["ping"].description, "Ping Command")
        self.assertEqual(self.cmd_tree.children["ping"].children, {})
        assert self.cmd_tree.children["ping"].parent is not None
        self.assertEqual(self.cmd_tree.children["ping"].parent.name, "/")

        self.cmd_tree.children["acs"].add_child(CmdTreeNode("acs_ctrl", "ACS Controller"))
        self.assertEqual(len(self.cmd_tree.children["acs"].children), 1)
        acs_ctrl_child = self.cmd_tree.children["acs"].children["acs_ctrl"]
        self.assertEqual(acs_ctrl_child.name, "acs_ctrl")
        self.assertEqual(acs_ctrl_child.description, "ACS Controller")
        assert acs_ctrl_child.parent is not None
        self.assertEqual(acs_ctrl_child.parent.name, "acs")

    def test_path_contained(self):
        self.base_tree()
        self.assertTrue(self.cmd_tree.contains_path("/"))

    def test_path_contained_invalid_input(self):
        self.assertFalse(self.cmd_tree.contains_path(""))

    def test_extract_node_invalid_input(self):
        self.assertIsNone(self.cmd_tree.extract_subnode(""))

    def test_extract_node_not_contained(self):
        self.tree_with_two_layers()
        self.assertIsNone(self.cmd_tree.extract_subnode("aocs"))

    def test_extract_subnode_simple(self):
        self.tree_with_two_layers()
        acs_node = self.cmd_tree.extract_subnode("acs")
        assert acs_node is not None
        self.assertEqual(acs_node.name, "acs")
        self.assertIsNotNone(acs_node.children["acs_ctrl"] is not None)

    def _generic_subnode_relativ_path_test(self, acs_ctrl_node: Optional[CmdTreeNode]):
        assert acs_ctrl_node is not None
        self.assertEqual(acs_ctrl_node.name, "acs_ctrl")
        self.assertIsNotNone(acs_ctrl_node.children["update_params"] is not None)

    def test_extract_subnode_relative_path(self):
        self.tree_with_two_layers()
        self.cmd_tree["acs"]["acs_ctrl"].add_child(
            CmdTreeNode("update_params", "Update Parameters")
        )
        self._generic_subnode_relativ_path_test(self.cmd_tree.extract_subnode("acs/acs_ctrl"))

    def test_extract_subnode_relativ_path_by_list(self):
        self.tree_with_two_layers()
        self.cmd_tree["acs"]["acs_ctrl"].add_child(
            CmdTreeNode("update_params", "Update Parameters")
        )
        self._generic_subnode_relativ_path_test(
            self.cmd_tree.extract_subnode_by_node_list(["acs", "acs_ctrl"])
        )

    def test_path_contained_acs(self):
        self.base_tree()
        self.assertTrue(self.cmd_tree.contains_path("/acs"))

    def test_path_contained_tcs(self):
        self.base_tree()
        self.assertTrue(self.cmd_tree.contains_path("/tcs"))

    def test_path_contained_acs_ctrl(self):
        self.base_tree()
        self.cmd_tree.children["acs"].add_child(CmdTreeNode("acs_ctrl", "ACS Controller"))
        self.assertTrue(self.cmd_tree.contains_path("/acs/acs_ctrl"))

    def test_named_dict(self):
        self.base_tree()
        self.cmd_tree.children["acs"].add_child(CmdTreeNode("acs_ctrl", "ACS Controller"))
        name_dict = self.cmd_tree.name_dict
        root_dict = name_dict.get("/")
        assert root_dict is not None
        assert "tcs" in root_dict
        assert root_dict.get("tcs") is None
        assert "acs" in root_dict
        acs_dict = root_dict.get("acs")
        assert acs_dict is not None
        assert "acs_ctrl" in acs_dict
        assert acs_dict.get("acs_ctrl") is None

    def test_printout_empty(self):
        self.assertEqual(str(self.cmd_tree), f"/{os.linesep}")

    def test_prinout_one_sublevel(self):
        self.base_tree()
        self.cmd_tree.add_child(CmdTreeNode("ping", "Ping Command"))
        print(str(self.cmd_tree))
        self.assertEqual(
            str(self.cmd_tree),
            (
                f"/{os.linesep}"
                f"├── acs{os.linesep}"
                f"├── tcs{os.linesep}"
                f"└── ping{os.linesep}"
            ),
        )

    def test_prinout_two_sublevels(self):
        self.tree_with_two_layers()
        print(self.cmd_tree)
        self.assertEqual(
            str(self.cmd_tree),
            (
                f"/{os.linesep}"
                f"├── acs{os.linesep}"
                f"│  └── acs_ctrl{os.linesep}"
                f"├── tcs{os.linesep}"
                f"└── ping{os.linesep}"
            ),
        )

    def test_printout_two_sublevels_one_cutoff(self):
        self.tree_with_two_layers()
        printout = self.cmd_tree.str_for_tree(False, 1)
        print(printout)
        self.assertEqual(
            printout,
            (
                f"/{os.linesep}"
                f"├── acs{os.linesep}"
                f"│  └── ... (cut-off, maximum depth 1){os.linesep}"
                f"├── tcs{os.linesep}"
                f"└── ping{os.linesep}"
            ),
        )

    def test_printout_3(self):
        self.base_tree()
        self.cmd_tree.children["acs"].add_child(CmdTreeNode("acs_ctrl", "ACS Controller"))
        self.cmd_tree.children["tcs"].add_child(CmdTreeNode("tcs_ctrl", "TCS Controller"))

        print(self.cmd_tree)
        self.assertEqual(
            str(self.cmd_tree),
            (
                f"/{os.linesep}"
                f"├── acs{os.linesep}"
                f"│  └── acs_ctrl{os.linesep}"
                f"└── tcs{os.linesep}"
                f"   └── tcs_ctrl{os.linesep}"
            ),
        )

    def test_printout_4(self):
        self.base_tree()
        self.cmd_tree.children["acs"].add_child(CmdTreeNode("acs_ctrl", "ACS Controller"))
        self.cmd_tree.children["acs"].add_child(CmdTreeNode("mgm_0", "MGM 0"))
        self.cmd_tree.children["acs"].children["mgm_0"].add_child(
            CmdTreeNode("update_cfg", "Update Configuration")
        )
        self.cmd_tree.children["tcs"].add_child(CmdTreeNode("tcs_ctrl", "TCS Controller"))
        self.cmd_tree.children["tcs"].add_child(CmdTreeNode("pt1000_0", "PT1000 0"))
        self.cmd_tree.add_child(CmdTreeNode("ping", "Ping Command"))

        print(self.cmd_tree)
        self.assertEqual(
            str(self.cmd_tree),
            (
                f"/{os.linesep}"
                f"├── acs{os.linesep}"
                f"│  ├── acs_ctrl{os.linesep}"
                f"│  └── mgm_0{os.linesep}"
                f"│     └── update_cfg{os.linesep}"
                f"├── tcs{os.linesep}"
                f"│  ├── tcs_ctrl{os.linesep}"
                f"│  └── pt1000_0{os.linesep}"
                f"└── ping{os.linesep}"
            ),
        )

    def test_printout_5(self):
        self.base_tree()
        self.cmd_tree.children["acs"].add_child(CmdTreeNode("acs_ctrl", "ACS Controller"))
        self.cmd_tree.children["acs"].add_child(CmdTreeNode("mgm_0", "MGM 0"))
        self.cmd_tree.children["acs"].children["mgm_0"].add_child(
            CmdTreeNode("update_cfg", "Update Configuration")
        )
        tcs_ctrl = CmdTreeNode("tcs_ctrl", "TCS Controller")
        tcs_ctrl.add_child(CmdTreeNode("set_param", "Set Parameter"))
        self.cmd_tree.children["tcs"].add_child(tcs_ctrl)
        pt1000_node = CmdTreeNode("pt1000_0", "PT1000 0")
        pt1000_node.add_child(CmdTreeNode("set_mode", "Set Mode"))
        self.cmd_tree.children["tcs"].add_child(pt1000_node)
        self.cmd_tree.add_child(CmdTreeNode("ping", "Ping Command"))

        print(self.cmd_tree)
        self.assertEqual(
            str(self.cmd_tree),
            (
                f"/{os.linesep}"
                f"├── acs{os.linesep}"
                f"│  ├── acs_ctrl{os.linesep}"
                f"│  └── mgm_0{os.linesep}"
                f"│     └── update_cfg{os.linesep}"
                f"├── tcs{os.linesep}"
                f"│  ├── tcs_ctrl{os.linesep}"
                f"│  │  └── set_param{os.linesep}"
                f"│  └── pt1000_0{os.linesep}"
                f"│     └── set_mode{os.linesep}"
                f"└── ping{os.linesep}"
            ),
        )

    def _build_tree_with_hidden_children(self):
        self.base_tree()
        self.cmd_tree.children["acs"].add_child(CmdTreeNode("acs_ctrl", "ACS Controller"))
        mgm_0_node = CmdTreeNode("mgm_0", "MGM 0", hide_children_for_print=True)
        self.cmd_tree.children["acs"].add_child(mgm_0_node)
        update_cfg = CmdTreeNode("update_cfg", "Update Configuration")
        self.cmd_tree.children["acs"]["mgm_0"].add_child(update_cfg)
        tcs_ctrl = CmdTreeNode("tcs_ctrl", "TCS Controller", hide_children_for_print=True)
        tcs_ctrl.add_child(CmdTreeNode("set_param", "Set Parameter"))
        self.cmd_tree.children["tcs"].add_child(tcs_ctrl)
        pt1000_node = CmdTreeNode("pt1000_0", "PT1000 0", hide_children_for_print=True)
        pt1000_node.add_child(CmdTreeNode("set_mode", "Set Mode"))
        pt1000_node.add_child(CmdTreeNode("set_cfg", "Set Config"))
        self.cmd_tree.children["tcs"].add_child(pt1000_node)
        self.cmd_tree.add_child(CmdTreeNode("ping", "Ping Command"))

    def test_printout_hidden_children(self):
        self._build_tree_with_hidden_children()
        print(self.cmd_tree)
        self.assertEqual(
            str(self.cmd_tree),
            (
                f"/{os.linesep}"
                f"├── acs{os.linesep}"
                f"│  ├── acs_ctrl{os.linesep}"
                f"│  └── mgm_0{os.linesep}"
                f"│     └── ... (cut-off, children are hidden){os.linesep}"
                f"├── tcs{os.linesep}"
                f"│  ├── tcs_ctrl{os.linesep}"
                f"│  │  └── ... (cut-off, children are hidden){os.linesep}"
                f"│  └── pt1000_0{os.linesep}"
                f"│     └── ... (cut-off, children are hidden){os.linesep}"
                f"└── ping{os.linesep}"
            ),
        )

    def test_printout_hidden_override(self):
        self._build_tree_with_hidden_children()
        print(self.cmd_tree)
        printout = self.cmd_tree.str_for_tree(
            with_description=False, max_depth=None, show_hidden_elements=True
        )
        print(printout)
        self.assertEqual(
            printout,
            (
                f"/{os.linesep}"
                f"├── acs{os.linesep}"
                f"│  ├── acs_ctrl{os.linesep}"
                f"│  └── mgm_0{os.linesep}"
                f"│     └── update_cfg{os.linesep}"
                f"├── tcs{os.linesep}"
                f"│  ├── tcs_ctrl{os.linesep}"
                f"│  │  └── set_param{os.linesep}"
                f"│  └── pt1000_0{os.linesep}"
                f"│     ├── set_mode{os.linesep}"
                f"│     └── set_cfg{os.linesep}"
                f"└── ping{os.linesep}"
            ),
        )

    def _build_tree_with_suppressed_leaves(self):
        self.cmd_tree.children["acs"].add_child(CmdTreeNode("acs_ctrl", "ACS Controller"))
        mgm_0_node = CmdTreeNode("mgm_0", "MGM 0")
        self.cmd_tree.children["acs"].add_child(mgm_0_node)
        update_cfg = CmdTreeNode("update_cfg", "Update Configuration")
        self.cmd_tree.children["acs"]["mgm_0"].add_child(update_cfg)
        tcs_ctrl = CmdTreeNode("tcs_ctrl", "TCS Controller", hide_children_for_print=True)
        tcs_ctrl.add_child(CmdTreeNode("set_param", "Set Parameter"))
        self.cmd_tree.children["tcs"].add_child(tcs_ctrl)
        pt1000_node = CmdTreeNode("pt1000_0", "PT1000 0", hide_children_for_print=True)
        pt1000_node.add_child(CmdTreeNode("set_mode", "Set Mode"))
        self.cmd_tree.children["tcs"].add_child(pt1000_node)
        self.cmd_tree.children["tcs"].add_child(CmdTreeNode("heaters", "Heaters"))
        self.cmd_tree.add_child(CmdTreeNode("ping", "Ping Command"))
        self.cmd_tree.children["acs"].hide_children_which_are_leaves = True
        self.cmd_tree.children["tcs"].hide_children_which_are_leaves = True
        self.cmd_tree["ping"].hide_children_which_are_leaves = True
        self.cmd_tree["ping"].add_child(CmdTreeNode("event", "Event Test"))
        self.cmd_tree["ping"]["event"].add_child(CmdTreeNode("0", "Event 0"))

    def test_printout_suppressed_leaves(self):
        self.base_tree()
        self._build_tree_with_suppressed_leaves()
        print(self.cmd_tree)
        self.assertEqual(
            str(self.cmd_tree),
            (
                f"/{os.linesep}"
                f"├── acs{os.linesep}"
                f"│  ├── mgm_0{os.linesep}"
                f"│  │  └── update_cfg{os.linesep}"
                f"│  └── ... (cut-off, leaves are hidden){os.linesep}"
                f"├── tcs{os.linesep}"
                f"│  ├── tcs_ctrl{os.linesep}"
                f"│  │  └── ... (cut-off, children are hidden){os.linesep}"
                f"│  ├── pt1000_0{os.linesep}"
                f"│  │  └── ... (cut-off, children are hidden){os.linesep}"
                f"│  └── ... (cut-off, leaves are hidden){os.linesep}"
                f"└── ping{os.linesep}"
                f"   └── event{os.linesep}"
                f"      └── 0{os.linesep}"
            ),
        )

    def test_printout_suppressed_leaves_print_override(self):
        self.base_tree()
        self._build_tree_with_suppressed_leaves()
        printout = self.cmd_tree.str_for_tree(False, max_depth=None, show_hidden_elements=True)
        print(printout)
        self.assertEqual(
            printout,
            (
                f"/{os.linesep}"
                f"├── acs{os.linesep}"
                f"│  ├── acs_ctrl{os.linesep}"
                f"│  └── mgm_0{os.linesep}"
                f"│     └── update_cfg{os.linesep}"
                f"├── tcs{os.linesep}"
                f"│  ├── tcs_ctrl{os.linesep}"
                f"│  │  └── set_param{os.linesep}"
                f"│  ├── pt1000_0{os.linesep}"
                f"│  │  └── set_mode{os.linesep}"
                f"│  └── heaters{os.linesep}"
                f"└── ping{os.linesep}"
                f"   └── event{os.linesep}"
                f"      └── 0{os.linesep}"
            ),
        )



from deprecated.sphinx import deprecated
import logging
from typing import Optional
from abc import abstractmethod, ABC

from prompt_toolkit.history import History

from tmtccmd.util.obj_id import ObjectIdDictT

from tmtccmd.config.tmtc import CmdTreeNode
from tmtccmd.core import BackendBase

from .com import ComCfgBase, ComInterface
from .tmtc import TmtcDefinitionWrapper
from .defs import default_json_path, CORE_COM_IF_DICT, ComIfMapping


_LOGGER = logging.getLogger(__name__)


class HookBase(ABC):
    """This hook allows users to adapt the TMTC commander core to the unique mission requirements.
    It is used by implementing all abstract functions and then passing the instance to the
    TMTC commander core.
    """

    def __init__(self, cfg_file_path: Optional[str] = None):
        self.cfg_path = cfg_file_path
        if self.cfg_path is None:
            self.cfg_path = default_json_path()

    @abstractmethod
    def get_communication_interface(self, com_if_key: str) -> Optional[ComInterface]:
        from tmtccmd.config.com import create_com_interface_default

        assert self.cfg_path is not None
        cfg_base = ComCfgBase(com_if_key=com_if_key, json_cfg_path=self.cfg_path)
        return create_com_interface_default(cfg_base)

    @abstractmethod
    def get_command_definitions(self) -> CmdTreeNode:
        """This function should return the root node of the command definition tree."""
        pass

    def get_com_if_dict(self) -> ComIfMapping:
        return CORE_COM_IF_DICT

    def get_cmd_history(self) -> Optional[History]:
        """Optionlly return a history class for the past command paths which will be used
        when prompting a command path from the user in CLI mode."""
        return None

    def perform_mode_operation(self, tmtc_backend: BackendBase, mode: int):
        """Perform custom mode operations.

        :param tmtc_backend:
        :param mode:
        :return:
        """
        _LOGGER.warning("No custom mode operation implemented")

    @deprecated(version="8.0.0", reason="application specific code")
    def get_object_ids(self) -> ObjectIdDictT:
        from tmtccmd.config.objects import get_base_component_id_mapping

        """The user can specify an object ID dictionary here mapping object ID bytearrays to a
        list. This list could contain containing the string representation or additional
        information about that object ID.
        """
        return get_base_component_id_mapping()

    @deprecated(
        version="8.0.0rc0",
        reason="implement and use get_command_definitions instead",
    )
    def get_tmtc_definitions(self) -> TmtcDefinitionWrapper:
        """This is a dicitonary mapping services represented by strings to an operation code
        dictionary.

        :return:
        """
        from tmtccmd.config.globals import get_default_tmtc_defs

        return get_default_tmtc_defs()

    @deprecated(
        version="8.0.0rc0",
        reason="implement and use get_communication_interface instead",
    )
    def assign_communication_interface(self, com_if_key: str) -> Optional[ComInterface]:
        """Assign the communication interface used by the TMTC commander to send and receive
        TMTC with.

        :param com_if_key:      String key of the communication interface to be created.
        """
        return self.get_communication_interface(com_if_key)



from typing import Any
import enum
from abc import abstractmethod


class FrontendBase:
    @abstractmethod
    def start(self, args: Any):
        """
        Start the frontend.
        :return:
        """


class TcMode(enum.IntEnum):
    IDLE = 0
    ONE_QUEUE = 1
    MULTI_QUEUE = 2


class TmMode(enum.IntEnum):
    IDLE = 0
    LISTENER = 1


class BackendRequest(enum.IntEnum):
    """These requests can be treated like recommendations on what to do after calling the backend
    handler functions and the :py:meth:`BackendState.mode_to_req` function.

    Brief explanation of fields:
     1. NONE: No special recommendation
     2. TERMINATION_NO_ERROR: Will be returned for the One Queue mode after finishing queue
        handling.
     3. DELAY_IDLE: TC and TM mode are idle, so there is nothing to do
     4. DELAY_LISTENER: TC handling is not active but TM listening is active. Delay to
        wait for new TM packets
     5. CALL_NEXT: It is recommended to call the handler functions immediately, for example to
        handle the next entry in the TC queue
    """

    NONE = 0
    TERMINATION_NO_ERROR = 1
    DELAY_IDLE = 2
    DELAY_LISTENER = 3
    DELAY_CUSTOM = 4
    CALL_NEXT = 5


class ModeWrapper:
    def __init__(self):
        self.tc_mode = TcMode.IDLE
        self.tm_mode = TmMode.IDLE

    def __str__(self):
        return f"{self.__class__.__name__}: tc_mode={self.tc_mode!r}," f" tm_mode={self.tm_mode!r}"



from unittest import TestCase

from spacepackets.ccsds.time import CdsShortTimestamp
from spacepackets.ecss.tm import PusTelemetry
from tmtccmd.pus.s200_fsfw_mode import Subservice


class TestSrv200Tm(TestCase):
    def setUp(self):
        self.app_data = bytearray()
        self.app_data.extend(bytes([0x01, 0x02, 0x03, 0x04]))
        self.srv_200_tm = PusTelemetry(
            service=200,
            subservice=Subservice.TM_MODE_REPLY,
            timestamp=CdsShortTimestamp.empty().pack(),
            apid=0x02,
        )



import enum
import abc
import logging
from typing import Deque, List, Any, Dict, Optional
from spacepackets.ecss.tm import PusTelemetry


TelemetryList = List[bytes]
# Deprecated type alias.
TelemetryListT = TelemetryList
TelemetryQueueT = Deque[bytes]

PusTmQueue = Deque[PusTelemetry]
PusTmListT = List[PusTelemetry]

PusTmQueueT = Deque[PusTmListT]


class SpecificApidHandlerBase(abc.ABC):
    """Abstract base class for an CCSDS APID specific handler. The user can implement a TM handler
    by implementing this class and then adding it to the :py:class:`CcsdsTmHandler`.
    If a CCSDS space packet with a specific APID is received, it will be routed to this handler
    using the :py:func:`handle_tm` callback function
    """

    def __init__(self, apid: int, user_args: Any):
        self.apid = apid
        self.user_args: Any = user_args

    @abc.abstractmethod
    def handle_tm(self, _packet: bytes, _user_args: Any):
        logging.getLogger(__name__).warning(f"No TM handling implemented for APID {self.apid}")


class GenericApidHandlerBase(abc.ABC):
    """This class is similar to the :py:class:`SpecificApidHandlerBase` but it is not specific
    for an APID and the found APID will be passed to the callback
    """

    def __init__(self, user_args: Any):
        self.user_args: Any = user_args

    @abc.abstractmethod
    def handle_tm(self, apid: int, _packet: bytes, _user_args: Any):
        pass


class DefaultApidHandler(GenericApidHandlerBase):
    def handle_tm(self, apid: int, _packet: bytes, _user_args: Any):
        logging.getLogger(__name__).warning(f"No TM handling implemented for unknown APID {apid}")


HandlerDictT = Dict[int, SpecificApidHandlerBase]


class TmTypes(enum.Enum):
    NONE = enum.auto
    CCSDS_SPACE_PACKETS = enum.auto


class TmHandlerBase:
    def __init__(self, tm_type: TmTypes):
        self._tm_type = tm_type

    def get_type(self):
        return self._tm_type


class CcsdsTmHandler(TmHandlerBase):
    """Generic CCSDS handler class. The user can create an instance of this class to handle
    CCSDS packets by adding dedicated APID handlers or a generic handler for all APIDs with no
    dedicated handler"""

    def __init__(self, generic_handler: Optional[GenericApidHandlerBase]):
        super().__init__(tm_type=TmTypes.CCSDS_SPACE_PACKETS)
        self._handler_dict: HandlerDictT = dict()
        if generic_handler is None:
            self.generic_handler = DefaultApidHandler(None)
        else:
            self.generic_handler = generic_handler

    def add_apid_handler(self, handler: SpecificApidHandlerBase):
        """Add a TM handler for a certain APID. The handler is a callback function which
        will be called if telemetry with that APID arrives.

        :param handler: Handler class instance
        :return:
        """
        self._handler_dict[handler.apid] = handler

    def has_apid(self, apid: int) -> bool:
        return apid in self._handler_dict

    def user_hook(self, apid: int, packet: bytes):
        """Can be overriden to trace all packets received."""
        pass

    def handle_packet(self, apid: int, packet: bytes) -> bool:
        """Handle a packet with an APID. If a handler exists for the given APID,
        it is used to handle the packet. If not, a dedicated handler for unknown APIDs
        is called.

        :param apid:
        :param packet:
        :return: True if the packet was passed to as dedicated APID handler, False otherwise
        """
        self.user_hook(apid, packet)
        specific_handler = self._handler_dict.get(apid)
        if specific_handler is None:
            self.generic_handler.handle_tm(apid, packet, self.generic_handler.user_args)
            return False
        specific_handler.handle_tm(packet, specific_handler.user_args)
        return True



"""Configuration helpers and definitions.

Submodules:

* :py:mod:`tmtccmd.config.hook` - Base hook class which should be implemented by user and is used
  by the framework to retrieve certain configuration from the user.
* :py:mod:`tmtccmd.config.args` - Various helper methods and classes to create the argument parsers
  and arguments converts to create the data structures expected by this library from passed CLI
  arguments."""

import logging
from pathlib import Path
from typing import Optional

from cfdppy.request import PutRequest
from spacepackets.cfdp import CfdpLv
from spacepackets.cfdp.tlv import ProxyPutRequest, ProxyPutRequestParams
from spacepackets.util import UnsignedByteField

from tmtccmd.cfdp.request import PutRequestCfgWrapper
from tmtccmd.core import TcMode, TmMode
from tmtccmd.core.base import ModeWrapper
from tmtccmd.tmtc.procedure import (
    CfdpProcedure,
    TreeCommandingProcedure,
    ProcedureWrapper,
    TcProcedureType,
)

from .args import (
    TreeCommandingParams,
    PreArgsParsingWrapper,
    ProcedureParamsWrapper,
    SetupParams,
    add_default_tmtccmd_args,
    create_default_args_parser,
    parse_default_tmtccmd_input_arguments,
)
from .defs import (
    CORE_COM_IF_DICT,
    CfdpParams,
    ComIfDictT,
    CoreComInterfaces,
    CoreModeConverter,
    CoreModeList,
    default_json_path,
)
from .hook import HookBase
from .prompt import prompt_op_code, prompt_service
from .tmtc import CmdTreeNode, OpCodeEntry, OpCodeOptionBase, TmtcDefinitionWrapper


def backend_mode_conversion(mode: str, mode_wrapper: ModeWrapper):
    if mode == CoreModeConverter.get_str(CoreModeList.LISTENER_MODE):
        mode_wrapper.tm_mode = TmMode.LISTENER
        mode_wrapper.tc_mode = TcMode.IDLE
    elif mode == CoreModeConverter.get_str(CoreModeList.ONE_QUEUE_MODE):
        mode_wrapper.tm_mode = TmMode.LISTENER
        mode_wrapper.tc_mode = TcMode.ONE_QUEUE
    elif mode == CoreModeConverter.get_str(CoreModeList.MULTI_INTERACTIVE_QUEUE_MODE):
        mode_wrapper.tc_mode = TcMode.MULTI_QUEUE
        mode_wrapper.tm_mode = TmMode.LISTENER


class SetupWrapper:
    """This class encapsulates various important setup parameters required by tmtccmd components"""

    def __init__(
        self,
        hook_obj: HookBase,
        setup_params: SetupParams,
        proc_param_wrapper: ProcedureParamsWrapper,
        json_cfg_path: Optional[str] = None,
    ):
        """
        :param hook_obj: User hook object. Needs to be implemented by the user
        :param setup_params: Optional helper wrapper which contains CLI arguments.
        """
        self.hook_obj = hook_obj
        self.json_cfg_path = json_cfg_path
        self._params = setup_params
        self.proc_param_wrapper = proc_param_wrapper
        self.json_cfg_path = json_cfg_path
        if json_cfg_path is None:
            self.json_cfg_path = default_json_path()

    @property
    def params(self):
        return self._params


def tmtc_params_to_procedure(params: TreeCommandingParams) -> TreeCommandingProcedure:
    return TreeCommandingProcedure(cmd_path=params.cmd_path)


def cfdp_put_req_params_to_procedure(params: CfdpParams) -> CfdpProcedure:
    proc_info = CfdpProcedure()
    proc_info.request_wrapper.base = PutRequestCfgWrapper(params)
    return proc_info


def params_to_procedure_conversion(
    param_wrapper: ProcedureParamsWrapper,
) -> ProcedureWrapper:
    proc_wrapper = ProcedureWrapper(None)
    if param_wrapper.ptype == TcProcedureType.TREE_COMMANDING:
        tree_cmd_params = param_wrapper.tree_commanding_params()
        assert tree_cmd_params is not None
        proc_wrapper.procedure = tmtc_params_to_procedure(tree_cmd_params)
    elif param_wrapper.ptype == TcProcedureType.CFDP:
        proc_wrapper.procedure = cfdp_put_req_params_to_procedure(
            param_wrapper.cfdp_params()  # type: ignore
        )
    return proc_wrapper
